from __future__ import annotations

import copy
import json
import warnings
import webbrowser
from contextlib import contextmanager
from typing import TYPE_CHECKING, Literal, Optional, Union

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

import neuroglancer
import numpy as np
import pyperclip
from IPython.display import HTML
from neuroglancer import viewer, viewer_base

from . import source_info
from .ngl_components import (
    AnnotationLayer,
    CoordSpace,
    DataMap,
    ImageLayer,
    RawLayer,
    SegmentationLayer,
)
from .site_utils import MAX_URL_LENGTH, neuroglancer_url
from .utils import NamedList, strip_layers, strip_state_properties

if TYPE_CHECKING:
    import caveclient
    import pandas as pd


class UnservedViewer(viewer_base.UnsynchronizedViewerBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._default_viewer_url = ""

    def get_server_url(self):
        return self._default_viewer_url


class ViewerState:
    def __init__(
        self,
        target_site: str = None,
        target_url: str = None,
        layers: Optional[list] = None,
        dimensions: Optional[Union[list, CoordSpace]] = None,
        *,
        position: Optional[Union[list, np.ndarray]] = None,
        scale_imagery: float = 1.0,
        scale_3d: float = 50000.0,
        show_slices: bool = False,
        selected_layer: Optional[str] = None,
        selected_layer_visible: bool = False,
        layout: Literal[
            "xy", "yz", "xz", "xy-3d", "xz-3d", "yz-3d", "4panel", "3d", "4panel-alt"
        ] = "xy-3d",
        base_state: Optional[dict] = None,
        interactive: bool = False,
        infer_coordinates: bool = True,
        client: Optional["caveclient.CAVEclient"] = None,
    ):
        """
        Parameters
        ----------
        target_site : str
            The target site for the viewer. If None, the default site will be used.
        target_url : str
            The target URL for the viewer. If None, the default URL will be used.
        layers : list of Layer
            The layers to add to the viewer.
        dimensions : list or CoordSpace
            The dimensions of the viewer. If None, the default dimensions will be used.
        position : list or np.ndarray
            The position of the viewer in 3D space. If None, the default position will be used.
        scale_imagery : float
            The scale factor for imagery layers. Default is 1.0.
        scale_3d : float
            The scale factor for 3D projections. Default is 50000.0.
        show_slices : bool
            Whether to show cross-sectional slices in the viewer. Default is False.
        selected_layer : str
            The name of the selected layer. If None, no layer is selected.
        selected_layer_visible : bool
            Whether the selected layer is visible. Default is False.
        layout : str
            The panel layout of the viewer. Default is "xy-3d".
        base_state : dict
            The base state of the viewer. If None, the default state will be used.
        interactive : bool
            Whether the viewer is interactive. Default is False.
        infer_coordinates : bool
            Whether to infer resolution and position from the source information using CloudVolume. Default is True.
        client : caveclient.CAVEclient, optional
            A CAVE client to use for configuration. If provided, it will be used by default in functions that can accept it.
        """

        self._target_site = target_site
        self._target_url = neuroglancer_url(target_url, target_site)
        self._layers = NamedList(layers) if layers else NamedList()
        self._dimensions = dimensions
        self._position = position
        self._scale_imagery = scale_imagery
        self._scale_3d = scale_3d
        self._show_slices = show_slices
        self._selected_layer = selected_layer
        self._selected_layer_visible = selected_layer_visible
        self._layout = layout
        self._base_state = base_state
        self._interactive = interactive
        self._saved_state_url = None
        self._viewer = None
        self._source_info = None
        self._infer_coordinates = infer_coordinates
        self._client = client

    def add_layers_from_client(
        self,
        client: Optional["caveclient.CAVEclient"] = None,
        imagery: Union[bool, str] = True,
        segmentation: Union[bool, str] = True,
        skeleton_source: bool = True,
        resolution: bool = True,
        target_url: bool = False,
        selected_alpha: Optional[float] = None,
        alpha_3d: Optional[float] = None,
        mesh_silhouette: Optional[float] = None,
        imagery_kws: Optional[dict] = None,
        segmentation_kws: Optional[dict] = None,
    ) -> Self:
        """Configure the viewer with information from a CaveClient.
        Can set the target URL, viewer resolution, and add image and segmentation layers.

        Parameters
        ----------
        client : caveclient.CAVEclient, optional
            The client to use for configuration. If None, will use the client set in the viewer state if allowed.
        imagery : Union[bool, str], optional
            Whether to add an image layer, by default True.
            If a string is provided, it will be used as the name of the layer.
        segmentation : Union[bool, str], optional
            Whether to add a segmentation layer, by default True
            If a string is provided, it will be used as the name of the layer.
        skeleton_source : bool, optional
            Whether to try to add a skeleton source for the segmentation layer, if provided, by default True
        resolution : bool, optional
            Whether to infer viewer resolution from the client info, by default True
        target_url : bool, optional
            Whether to set the neuroglancer URL from the client, by default False
        selected_alpha : Optional[float], optional
            The alpha value for the segmentation layer, by default None.
        alpha_3d : Optional[float], optional
            The alpha value for 3D meshes, by default None.
        mesh_silhouette : Optional[float], optional
            The mesh silhouette value, by default None.
        imagery_kws : Optional[dict], optional
            Additional keyword arguments to pass to the image layer constructor.
        segmentation_kws : Optional[dict], optional
            Additional keyword arguments to pass to the segmentation layer constructor.

        Returns
        -------
        Self
            The updated viewer state object.
        """
        if client is None:
            client = self._client
        if client is None:
            raise ValueError(
                "No client provided and no client set in the viewer state. "
                "Please provide a CAVEclient instance."
            )

        if target_url:
            self._target_url = client.info.viewer_site()
        if resolution:
            self.dimensions = client.info.viewer_resolution()
        if imagery:
            if isinstance(imagery, str):
                self.add_image_layer(
                    source=client.info.image_source(),
                    name=imagery,
                    **(imagery_kws or {}),
                )
            else:
                self.add_image_layer(
                    source=client.info.image_source(),
                    **(imagery_kws or {}),
                )
        if segmentation:
            seg_source = [client.info.segmentation_source()]
            skel_source_path = client.info.get_datastack_info().get("skeleton_source")

            if skeleton_source and skel_source_path:
                seg_source.append(skel_source_path)

            if isinstance(segmentation, str):
                self.add_segmentation_layer(
                    source=seg_source,
                    name=segmentation,
                    selected_alpha=selected_alpha,
                    alpha_3d=alpha_3d,
                    mesh_silhouette=mesh_silhouette,
                    **(segmentation_kws or {}),
                )
            else:
                self.add_segmentation_layer(
                    source=seg_source,
                    selected_alpha=selected_alpha,
                    alpha_3d=alpha_3d,
                    mesh_silhouette=mesh_silhouette,
                    **(segmentation_kws or {}),
                )
        return self

    @property
    def viewer(self):
        """Get the Neuroglancer viewer object."""
        if self._viewer is None:
            self._viewer = self.to_neuroglancer_state()
        return self._viewer

    def _reset_viewer(self):
        self._viewer = None

    @property
    def infer_coordinates(self):
        return self._infer_coordinates

    @infer_coordinates.setter
    def infer_coordinates(self, value: bool):
        if value != self._infer_coordinates:
            self._reset_viewer()
        self._infer_coordinates = value

    @property
    def layers(self):
        return self._layers

    @property
    def layer_names(self):
        return [layer.name for layer in self.layers]

    def add_layer(self, layers, selected: bool = False) -> Self:
        """Add a layer to the viewer.
        Parameters
        ----------
        layers : Layer or list of Layer
            The layer(s) to add to the viewer.
        """

        if isinstance(layers, list):
            self._layers.extend(layers)
            if selected is not False:
                for layer, vis in zip(layers, selected):
                    if vis:
                        self.set_selected_layer(layer.name)
        else:
            self._layers.append(layers)
            if selected:
                self.set_selected_layer(layers.name)
        self._reset_viewer()
        return self

    @property
    def source_info(self):
        if self._source_info is None:
            self._source_info = source_info.populate_info(self.layers)
        return self._source_info

    def _suggest_position_from_source(self, resolution=None) -> np.ndarray:
        """Suggest a position based on the source information.
        This uses the headers on each layer source to suggest a position based on the
        center of the first segmentation layer or, if no segmentation layer, the first image layer.
        If no position can be inferred, the position will be to None.

        NOTE: Requires cloudvolume to be installed and an internet connection.

        Parameters
        ----------
        resolution : list or np.ndarray, optional
            The resolution of the viewer. If not provided, the viewer's current resolution will be used.

        Returns
        -------
        Position : np.ndarray

        """
        position = source_info.suggest_position(self.source_info, resolution)
        return position

    def _suggest_resolution_from_source(self) -> np.ndarray:
        """Suggest a resolution based on the source information.
        This uses the headers on each layer source to suggest a resolution based on the
        first image layer. If no resolution can be inferred, the resolution will be None.

        NOTE: Requires cloudvolume to be installed and an internet connection.

        Returns
        -------
        Resolution : np.ndarray

        """
        resolution = source_info.suggest_resolution(self.source_info)
        return resolution

    @property
    def dimensions(self):
        return self._dimensions

    @dimensions.setter
    def dimensions(self, value):
        if isinstance(value, CoordSpace):
            self._dimensions = value
        else:
            self._dimensions = CoordSpace(value)
        self._reset_viewer()

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, value):
        self._position = value
        self._reset_viewer()

    @property
    def scale_imagery(self):
        return self._scale_imagery

    @scale_imagery.setter
    def scale_imagery(self, value):
        self._scale_imagery = value
        self._reset_viewer()

    @property
    def scale_3d(self):
        return self._scale_3d

    @scale_3d.setter
    def scale_3d(self, value):
        self._scale_3d = value
        self._viewer = None

    @property
    def show_slices(self):
        return self._show_slices

    @show_slices.setter
    def show_slices(self, value):
        self._show_slices = value
        self._viewer = None

    @property
    def selected_layer(self):
        return {
            "layer": self._selected_layer,
            "visible": self._selected_layer_visible,
        }

    @selected_layer.setter
    def selected_layer(
        self, value: Union[str, ImageLayer, SegmentationLayer, AnnotationLayer]
    ):
        self.set_selected_layer(value)

    def set_selected_layer(
        self, selected_layer: Union[str, ImageLayer, SegmentationLayer, AnnotationLayer]
    ) -> Self:
        if isinstance(selected_layer, str):
            self._selected_layer = selected_layer
        else:
            self._selected_layer = selected_layer.name
        self._reset_viewer()
        return self

    @property
    def selected_layer_visible(self):
        return self._selected_layer_visible

    @property
    def layout(self):
        return self._layout

    @layout.setter
    def layout(self, value):
        if value not in [
            "xy",
            "yz",
            "xz",
            "xy-3d",
            "xz-3d",
            "yz-3d",
            "4panel",
            "3d",
            "4panel-alt",
        ]:
            raise ValueError(
                f"Invalid layout: {value}. Must be one of 'xy', 'yz', 'xz', 'xy-3d', 'xz-3d', 'yz-3d', '4panel', '3d', or '4panel-alt'."
            )
        self._layout = value
        self._reset_viewer()

    @property
    def base_state(self):
        return self._base_state

    @base_state.setter
    def base_state(self, value):
        self._base_state = value
        self._reset_viewer()

    @property
    def interactive(self):
        return self._interactive

    @interactive.setter
    def interactive(self, value):
        if value != self._interactive:
            self._viewer = None
        self._interactive = value

    def set_viewer_properties(
        self,
        position: Optional[Union[list, np.ndarray]] = None,
        dimensions: Optional[Union[list, np.ndarray]] = None,
        scale_imagery: Optional[float] = None,
        scale_3d: Optional[float] = None,
        show_slices: Optional[bool] = None,
        selected_layer: Optional[Union[str, ImageLayer]] = None,
        selected_layer_visible: Optional[bool] = None,
        layout: Optional[
            Literal[
                "xy",
                "yz",
                "xz",
                "xy-3d",
                "xz-3d",
                "yz-3d",
                "4panel",
                "3d",
                "4panel-alt",
            ]
        ] = None,
        base_state: Optional[dict] = None,
        interactive: Optional[bool] = None,
        infer_coordinates: bool = False,
    ) -> Self:
        """
        Set various properties of the viewer state.

        This function allows you to configure the viewer's position, dimensions,
        scale, layout, and other properties. Any property not explicitly set will
        remain unchanged.

        Parameters
        ----------
        position : list or np.ndarray, optional
            The position of the viewer in 3D space.
        dimensions : list or np.ndarray, optional
            The dimensions of the viewer's coordinate space.
            Can be either a list of 3 values treated as nanometer resolution or
            a CoordSpace object with more detailed options.
        scale_imagery : float, optional
            The scale factor for imagery layers.
        scale_3d : float, optional
            The scale factor for 3D projections.
        show_slices : bool, optional
            Whether to show cross-sectional slices in the viewer.
        selected_layer : str or ImageLayer, optional
            The name of the selected layer or the layer object itself.
        selected_layer_visible : bool, optional
            Whether the selected layer is visible.
        layout : {"xy", "yz", "xz", "xy-3d", "xz-3d", "yz-3d", "4panel", "3d", "4panel-alt"}, optional
            The panel layout of the viewer.
        base_state : dict, optional
            The base state of the viewer.
        interactive : bool, optional
            Whether the viewer is interactive.
        infer_coordinates : bool, optional
            Whether to infer resolution and position from the source information.
            If True, the viewer will attempt to extract this information from the
            info provided by the source if not provided explicitly.


        Returns
        -------
        self
            The updated viewer state object.
        """

        if position is not None:
            self.position = position
        if dimensions is not None:
            self.dimensions = dimensions
        if scale_imagery is not None:
            self.scale_imagery = scale_imagery
        if scale_3d is not None:
            self.scale_3d = scale_3d
        if show_slices is not None:
            self.show_slices = show_slices
        if selected_layer is not None:
            if isinstance(selected_layer, str):
                self.selected_layer = selected_layer
            else:
                self.selected_layer = selected_layer.name
        if selected_layer_visible is not None:
            self.selected_layer_visible = selected_layer_visible
        if layout is not None:
            self.layout = layout
        if base_state is not None:
            self.base_state = base_state
        if interactive is not None:
            self.interactive = interactive
        if infer_coordinates is not None:
            self.infer_coordinates = infer_coordinates
        return self

    def add_raw_layer(
        self,
        data: dict,
        name: Optional[str] = None,
        visible: Optional[bool] = None,
        archived: Optional[bool] = None,
        source_remap: Optional[dict] = None,
    ) -> Self:
        """Add a layer based on a raw state dictionary. Useful when copying existing layers, but offers limited customization.

        Parameters
        ----------
        data : dict
            The raw state dictionary for the layer.
        name : str, optional
            The new name of the layer, if you want to change it.
        visible : bool, optional
            Whether the layer is visible, if you want to change it.
        archived : bool, optional
            Whether the layer is archived, if you want to change it.
        source_remap: dict, optional
            A dictionary mapping source cloudpaths to new cloudpaths, for example to update the layer's data source without changing anything else.

        Returns
        -------
        Self
            The updated viewer state object.
        """
        layer = RawLayer(
            json_data=data, name=name, visible=visible, archived=archived
        ).remap_sources(source_remap)
        self.add_layer(layer)
        return self

    def add_image_layer(
        self,
        source: str,
        name: str = "imagery",
        resolution: Optional[Union[list, np.ndarray]] = None,
        **kwargs,
    ) -> Self:
        """Add an image layer to the viewer.

        Parameters
        ----------
        source : str
            The source path for the image layer.
        name : str, optional
            The name of the image layer, by default "imagery".
        resolution : Optional[Union[list, np.ndarray]], optional
            The resolution of the image layer. If None, the viewer's current resolution will be used or it will be inferred from the source (if cloud-volume is installed).
        **kwargs : dict, optional
            Additional keyword arguments to pass to the image layer constructor.
        """
        if resolution is None:
            resolution = self.dimensions
        img_layer = ImageLayer(
            name=name,
            source=source,
            resolution=resolution,
            **kwargs,
        )
        self.add_layer(img_layer)
        return self

    def add_segmentation_layer(
        self,
        source: str,
        name: str = "segmentation",
        resolution: Optional[Union[list, np.ndarray]] = None,
        segments: Optional[list] = None,
        **kwargs,
    ) -> Self:
        """Add a segmentation layer to the viewer.

        Parameters
        ----------
        source : str
            The source path for the segmentation layer.
        name : str, optional
            The name of the segmentation layer, by default "segmentation".
        resolution : Optional[Union[list, np.ndarray]], optional
            The resolution of the segmentation layer. If None, the viewer's current resolution will be used or it will be inferred from the source (if cloud-volume is installed).
        segments: list, optional
            A list of segments to set as selected in the segmentation layer.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the segmentation layer constructor.
        """
        if resolution is None:
            resolution = self.dimensions
        seg_layer = SegmentationLayer(
            name=name,
            source=source,
            resolution=resolution,
            segments=segments,
            **kwargs,
        )
        self.add_layer(seg_layer)
        return self

    def add_segments(
        self,
        segments: Union[list, np.ndarray, "pd.DataFrame", DataMap],
        visible: Optional[Union[list, np.ndarray]] = None,
        segment_colors: Optional[Union[str, list, dict]] = None,
        name: Optional[str] = None,
    ) -> Self:
        """Add segments directly to an existing segmentation layer.
        By default, it will use the first segmentation layer found in the viewer, otherwise you must specify a layer name.

        Parameters
        ----------
        segments : list or dict or VisibleSegments
            The segments to add. If a dict, the keys are the segment IDs and the values are the boolean visibility.
        visible: list, optional
            The visibility of the segments, assumed to be True if not provided.
            Should be the same length as segments, and segments should be a list of the same length.
        segment_colors : Union[str, list, dict], optional
            The color(s) to assign to the segments. If a string or list, all segments will be assigned the same color.
            A list is assumed to be a color tuple.
            If a dict, the keys are segment IDs and the values are colors.
        name: Optional[str]
            The name of the segmentation layer to add segments to.
            If None, it will use the first segmentation layer found in the viewer.

        Returns
        -------
        The viewer state object with the added segments.
        """
        if name is None:
            for l in self.layers:
                if isinstance(l, SegmentationLayer):
                    name = l.name
                    break
            else:
                raise ValueError("No segmentation layer found in the viewer.")
        self.layers[name].add_segments(
            segments,
            visible=visible,
        )
        if segment_colors is not None:
            if not isinstance(segment_colors, dict):
                segment_colors = {s: segment_colors for s in segments}
            self.layers[name].add_segment_colors(segment_colors)
        return self

    def add_segment_properties(
        self,
        data: Union[pd.DataFrame, DataMap],
        id_column: str = "pt_root_id",
        client: Optional["caveclient.CAVEclient"] = None,
        label_column: Optional[Union[str, list]] = None,
        description_column: Optional[str] = None,
        string_columns: Optional[list] = None,
        number_columns: Optional[list] = None,
        tag_value_columns: Optional[list] = None,
        tag_bool_columns: Optional[list] = None,
        tag_descriptions: Optional[list] = None,
        allow_disambiguation: bool = True,
        label_separator: str = "_",
        label_format_map: Optional[str] = None,
        prepend_col_name: bool = False,
        random_columns: Optional[int] = None,
        random_column_prefix: str = None,
        dry_run: bool = False,
        name: Optional[str] = None,
    ) -> Self:
        """Upload segment properties and add to the layer.
        If you already have a segment properties cloud path, use `add_source` to add it to the layer.

        Parameters
        ----------
        data : pd.DataFrame
            The DataFrame containing the segment properties.
        id_column : str, optional
            The column name for the segment IDs. Default is 'pt_root_id'.
        client : CAVEclient
            The CAVEclient object needed to upload .
        label_column : str or list, optional
            The column name(s) for the segment labels. Default is None.
        description_column : str, optional
            The column name for the segment descriptions. Default is None.
        string_columns : list, optional
            The column names for string properties. Default is None.
        number_columns : list, optional
            The column names for number properties. Default is None.
        tag_value_columns : list, optional
            The column names for tag value properties. Default is None.
        tag_bool_columns : list, optional
            The column names for tag boolean properties. Default is None.
        tag_descriptions : list, optional
            The descriptions for the tags. Default is None.
        allow_disambiguation : bool, optional
            Whether to allow disambiguation of segment IDs. Default is True.
        label_separator : str, optional
            The separator for label formatting. Default is "_".
        label_format_map : str, optional
            The format map for labels. Default is None.
        prepend_col_name: bool, optional
            Whether to prepend the column name to the label. Default is False.
        random_columns: int, optional
            Number of random columns to add. Default is None.
        random_column_prefix: str, optional
            Name prefix of the random columns. Default is None.
        dry_run: bool, optional
            If dry run is true, build but do not actually upload and instead use a placeholder source text. Default is False.
        name: str, optional
            Layer name. Must be specified if there is more than one segmentation layer in the state.

        Returns
        -------
        Self
        """

        if name is None:
            for l in self.layers:
                if isinstance(l, SegmentationLayer):
                    name = l.name
                    break
            else:
                raise ValueError("No segmentation layer found in the viewer.")

        if client is None and self._client is not None:
            client = self._client
        else:
            raise ValueError("Client must be specified.")

        self.layers[name].add_segment_properties(
            data,
            id_column=id_column,
            client=client,
            label_column=label_column,
            description_column=description_column,
            string_columns=string_columns,
            number_columns=number_columns,
            tag_value_columns=tag_value_columns,
            tag_bool_columns=tag_bool_columns,
            tag_descriptions=tag_descriptions,
            allow_disambiguation=allow_disambiguation,
            label_separator=label_separator,
            label_format_map=label_format_map,
            prepend_col_name=prepend_col_name,
            random_columns=random_columns,
            random_column_prefix=random_column_prefix,
            dry_run=dry_run,
        )

        return self

    def add_segments_from_data(
        self,
        data: Union["pd.DataFrame", DataMap],
        segment_column: str,
        visible_column: str = None,
        color_column: str = None,
        name: Optional[str] = None,
    ) -> Self:
        """Add segments from a DataFrame or DataMap to an existing segmentation layer.
        By default, it will use the first segmentation layer found in the viewer, otherwise you must specify a layer name.

        Parameters
        ----------
        data : Union[pd.DataFrame, DataMap]
            The data containing segment information or a datamap to be filled in later
        segment_column : str
            The name of the column containing segment IDs in the DataFrame.
        visible_column : str, optional
            The name of the column containing visibility information in the DataFrame.
            If not provided, all segments will be considered visible.
        color_column : str, optional
            The name of the column containing color information for the segments.
        name: Optional[str]
            The name of the segmentation layer to add segments to.
            If None, it will use the first segmentation layer found in the viewer.
        """

        if name is None:
            for l in self.layers:
                if isinstance(l, SegmentationLayer):
                    name = l.name
                    break
            else:
                raise ValueError("No segmentation layer found in the viewer.")
        self.layers[name].add_segments_from_data(
            data,
            segment_column=segment_column,
            visible_column=visible_column,
            color_column=color_column,
        )
        return self

    def add_annotation_layer(
        self,
        name: str = "annotation",
        source: Optional[str] = None,
        resolution: Optional[Union[list, np.ndarray]] = None,
        tags: Optional[list] = None,
        linked_segmentation: Union[str, bool, dict] = True,
        shader: str = None,
        **kwargs,
    ) -> Self:
        """Add an annotation layer, either local or with precomputed annotation source

        Parameters
        ----------
        name : str, optional
            Layer name, by default "annotation"
        source : Optional[str], optional
            Cloud annotation source path, if using. By default None
        resolution : Optional[Union[list, np.ndarray]], optional
            Layer resolution, only used by local annotation layers.
            By default None, where it is inferred from the viewer.
        tags : Optional[list], optional
            Ordered list of tags, by default None.
        linked_segmentation : Union[str, bool], optional
            If True, will link to the first segmentation layer found in the viewer.
            If False, will not link to any segmentation layer.
            If a string is provided, it will be used as the name of the segmentation layer to link to.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the annotation layer constructor.

        Returns
        -------
        Self
            The viewer state object with the added annotation layer.
        """
        if linked_segmentation:
            if linked_segmentation is True:
                for layer in self.layers:
                    if isinstance(layer, SegmentationLayer):
                        linked_segmentation = layer.name
                        break
                else:
                    linked_segmentation = None
        if resolution is None:
            resolution = self.dimensions
        anno_layer = AnnotationLayer(
            name=name,
            source=source,
            resolution=resolution,
            tags=tags,
            linked_segmentation=linked_segmentation,
            shader=shader,
            **kwargs,
        )
        self.add_layer(anno_layer)
        return self

    def get_layer(self, name: str):
        """Get a layer by name.

        Parameters
        ----------
        name : str
            The name of the layer to get.

        Returns
        -------
        Layer
            The layer with the specified name.
        """
        for layer in self.layers:
            if layer.name == name:
                return layer
        raise ValueError(f"Layer {name} not found in viewer.")

    @contextmanager
    def with_datamap(self, datamap: dict):
        """Context manager to apply datamaps."""
        if datamap is None:
            yield self
        else:
            if not isinstance(datamap, dict):
                datamap = {None: datamap}
            viewer_copy = copy.deepcopy(self)
            viewer_copy.map(datamap, inplace=True)
            yield viewer_copy

    def _apply_datamaps(self, datamap: dict):
        for layer in self.layers:
            layer._apply_datamaps(datamap)

    def add_annotation_source(
        self,
        source: str,
        name: str = "annotation",
        linked_segmentation: Union[str, bool] = True,
        shader: Optional[str] = None,
    ) -> Self:
        """Add a precomputed annotation source to the viewer.

        Parameters
        ----------
        source : str
            The source path for the annotation layer.
        name : str, optional
            The name of the annotation layer, by default "annotation".
        shader : Optional[str], optional
            The shader to use for the annotation layer, by default None.

        Returns
        -------
        Self
            The viewer state object with the added annotation layer.
        """
        self.add_annotation_layer(
            source=source,
            name=name,
            linked_segmentation=linked_segmentation,
            shader=shader,
        )
        return self

    def add_points(
        self,
        data: Optional[Union[list, np.ndarray, "pd.DataFrame", DataMap]] = None,
        name: str = "annotation",
        point_column: Optional[str] = None,
        segment_column: Optional[str] = None,
        description_column: Optional[str] = None,
        tag_column: Optional[str] = None,
        tag_bools: Optional[list] = None,
        data_resolution: Optional[list] = None,
        tags: Optional[list] = None,
        linked_segmentation: Union[str, bool] = True,
        shader: Optional[str] = None,
        color: Optional[str] = None,
        swap_visible_segments_on_move: bool = True,
    ) -> Self:
        """Add points to an existing annotation layer or create a new one.
        Parameters
        ----------
        data : Union[list, np.ndarray, pd.DataFrame]
            The data to add to the annotation layer.
        name : str, optional
            The name of the annotation layer, by default "annotation".
        point_column : Optional[str], optional
            The name of the column containing point coordinates, by default None.
            None is needed if the data is an array, but is required if the data is a DataFrame.
        segment_column : Optional[str], optional
            The name of the column containing linked segment IDs, by default None.
        description_column : Optional[str], optional
            The name of the column containing descriptions, by default None.
            If None, no descriptions will be added.
        tag_column : Optional[str], optional
            The name of a column containing tags, by default None.
        tag_bools : Optional[list], optional
            A list of column names indicating tags as booleans, by default None.
        data_resolution : Optional[list], optional
            The resolution of the data, by default None.
            If None, the viewer's current resolution will be used.
        tags : Optional[list], optional
            A list of tags to add to the annotation layer, by default None.
            If None, tags will be inferred from the data, if any tag columns are provided.
        linked_segmentation : str or bool, optional
            The name of the segmentation layer to link to, by default None.
            If True, will link to the first segmentation layer found in the viewer.
            If False, will not link to any segmentation layer.
        shader : Optional[str], optional
            The shader to use for the annotation layer, by default None.
            If None, the default shader will be used.
        swap_visible_segments_on_move: bool, optional
            If True, will swap the visibility of segments when moving points.
        Returns
        -------
        Self
            The viewer state object with the added points.
        """

        if name in self.layer_names:
            layer = self.get_layer(name)
            if not isinstance(layer, AnnotationLayer):
                raise ValueError(
                    f"Layer {name} already exists but is not a AnnotationLayer."
                )
        else:
            layer = AnnotationLayer(
                name=name,
                resolution=data_resolution,
                tags=tags,
                linked_segmentation=linked_segmentation,
                color=color,
                shader=shader,
                swap_visible_segments_on_move=swap_visible_segments_on_move,
            ).add_points(
                data,
                point_column=point_column,
                segment_column=segment_column,
                description_column=description_column,
                tag_column=tag_column,
                tag_bools=tag_bools,
                data_resolution=data_resolution,
            )
            self.add_layer(layer)
        return self

    def add_lines(
        self,
        data: Optional["pd.DataFrame"] = None,
        name: str = "annotation",
        point_a_column: str = None,
        point_b_column: str = None,
        segment_column: Optional[str] = None,
        description_column: Optional[str] = None,
        tag_column: Optional[str] = None,
        tag_bools: Optional[list] = None,
        data_resolution: Optional[list] = None,
        tags: Optional[list] = None,
        linked_segmentation: Union[str, bool] = True,
        shader: Optional[str] = None,
        color: Optional[str] = None,
        swap_visible_segments_on_move: bool = True,
    ) -> Self:
        """Add lines to an existing annotation layer or create a new one.

        Parameters
        ----------
        data : pd.DataFrame
            The DataFrame containing the line annotations.
        name : str, optional
            The name of the annotation layer, by default "annotation".
        point_a_column : str
            The column name for the start point coordinates.
        point_b_column : str
            The column name for the end point coordinates.
        segment_column : Optional[str], optional
            The name of the column containing linked segment IDs, by default None.
        description_column : Optional[str], optional
            The name of the column containing descriptions, by default None.
        tag_column : Optional[str], optional
            The name of a column containing tags, by default None.
        tag_bools : Optional[list], optional
            A list of column names indicating tags as booleans, by default None.
        data_resolution : Optional[list], optional
            The resolution of the data, by default None.
        tags : Optional[list], optional
            A list of tags to add to the annotation layer, by default None.
        linked_segmentation : str or bool, optional
            The name of the segmentation layer to link to, by default None.
            If True, will link to the first segmentation layer found in the viewer.
            If False, will not link to any segmentation layer.
        shader : Optional[str], optional
            The shader to use for the annotation layer, by default None.
        color : Optional[str], optional
            The color to use for the lines, by default None.
        swap_visible_segments_on_move : bool, optional
            If True, will swap the visibility of segments when moving lines.
        Returns
        -------
        Self
            The viewer state object with the added lines.
        """
        if name in self.layer_names:
            layer = self.get_layer(name)
            if not isinstance(layer, AnnotationLayer):
                raise ValueError(
                    f"Layer {name} already exists but is not a AnnotationLayer."
                )
        else:
            layer = AnnotationLayer(
                name=name,
                resolution=data_resolution,
                tags=tags,
                linked_segmentation=linked_segmentation,
                shader=shader,
                color=color,
                swap_visible_segments_on_move=swap_visible_segments_on_move,
            ).add_lines(
                data,
                point_a_column=point_a_column,
                point_b_column=point_b_column,
                segment_column=segment_column,
                description_column=description_column,
                tag_column=tag_column,
                tag_bools=tag_bools,
                data_resolution=data_resolution,
            )
            self.add_layer(layer)
        return self

    def add_ellipsoids(
        self,
        data: Optional["pd.DataFrame"] = None,
        name: str = "annotation",
        center_column: str = None,
        radii_column: str = None,
        segment_column: Optional[str] = None,
        description_column: Optional[str] = None,
        tag_column: Optional[str] = None,
        tag_bools: Optional[list] = None,
        data_resolution: Optional[list] = None,
        tags: Optional[list] = None,
        linked_segmentation: Union[str, bool] = True,
        shader: Optional[str] = None,
        color: Optional[str] = None,
        swap_visible_segments_on_move: bool = True,
    ) -> Self:
        """Add ellipsoid annotations to an existing annotation layer or create a new one.

        Parameters
        ----------
        data : pd.DataFrame
            The DataFrame containing the ellipsoid annotations.
        name : str, optional
            The name of the annotation layer, by default "annotation".
        center_column : str
            The column name for the ellipsoid center coordinates.
        radii_column : str
            The column name for the ellipsoid radii.
        segment_column : Optional[str], optional
            The name of the column containing linked segment IDs, by default None.
        description_column : Optional[str], optional
            The name of the column containing descriptions, by default None.
        tag_column : Optional[str], optional
            The name of a column containing tags, by default None.
        tag_bools : Optional[list], optional
            A list of column names indicating tags as booleans, by default None.
        data_resolution : Optional[list], optional
            The resolution of the data, by default None.
        tags : Optional[list], optional
            A list of tags to add to the annotation layer, by default None.
        linked_segmentation : str or bool, optional
            The name of the segmentation layer to link to, by default None.
            If True, will link to the first segmentation layer found in the viewer.
            If False, will not link to any segmentation layer.
        shader : Optional[str], optional
            The shader to use for the annotation layer, by default None.
        color : Optional[str], optional
            The color to use for the ellipsoids, by default None.
        swap_visible_segments_on_move : bool, optional
            If True, will swap the visibility of segments when moving ellipsoids.
        Returns
        -------
        Self
            The viewer state object with the added ellipsoids.
        """
        if name in self.layer_names:
            layer = self.get_layer(name)
            if not isinstance(layer, AnnotationLayer):
                raise ValueError(
                    f"Layer {name} already exists but is not a AnnotationLayer."
                )
        else:
            layer = AnnotationLayer(
                name=name,
                resolution=data_resolution,
                tags=tags,
                linked_segmentation=linked_segmentation,
                shader=shader,
                color=color,
                swap_visible_segments_on_move=swap_visible_segments_on_move,
            ).add_ellipsoids(
                data,
                center_column=center_column,
                radii_column=radii_column,
                segment_column=segment_column,
                description_column=description_column,
                tag_column=tag_column,
                tag_bools=tag_bools,
                data_resolution=data_resolution,
            )
            self.add_layer(layer)
        return self

    def add_boxes(
        self,
        data: Optional["pd.DataFrame"] = None,
        name: str = "annotation",
        point_a_column: str = None,
        point_b_column: str = None,
        segment_column: Optional[str] = None,
        description_column: Optional[str] = None,
        tag_column: Optional[str] = None,
        tag_bools: Optional[list] = None,
        data_resolution: Optional[list] = None,
        tags: Optional[list] = None,
        linked_segmentation: Union[str, bool] = True,
        shader: Optional[str] = None,
        color: Optional[str] = None,
        swap_visible_segments_on_move: bool = True,
    ) -> Self:
        """Add bounding box annotations to an existing annotation layer or create a new one.

        Parameters
        ----------
        data : pd.DataFrame
            The DataFrame containing the bounding box annotations.
        name : str, optional
            The name of the annotation layer, by default "annotation".
        point_a_column : str
            The column name for the start point coordinates of the box.
        point_b_column : str
            The column name for the end point coordinates of the box.
        segment_column : Optional[str], optional
            The name of the column containing linked segment IDs, by default None.
        description_column : Optional[str], optional
            The name of the column containing descriptions, by default None.
        tag_column : Optional[str], optional
            The name of a column containing tags, by default None.
        tag_bools : Optional[list], optional
            A list of column names indicating tags as booleans, by default None.
        data_resolution : Optional[list], optional
            The resolution of the data, by default None.
        tags : Optional[list], optional
            A list of tags to add to the annotation layer, by default None.
        linked_segmentation : str or bool, optional
            The name of the segmentation layer to link to, by default None.
            If True, will link to the first segmentation layer found in the viewer.
            If False, will not link to any segmentation layer.
        shader : Optional[str], optional
            The shader to use for the annotation layer, by default None.
        color : Optional[str], optional
            The color to use for the bounding boxes, by default None.
        swap_visible_segments_on_move : bool, optional
            If True, will swap the visibility of segments when moving boxes.

        Returns
        -------
        Self
            The viewer state object with the added bounding boxes.
        """
        if name in self.layer_names:
            layer = self.get_layer(name)
            if not isinstance(layer, AnnotationLayer):
                raise ValueError(
                    f"Layer {name} already exists but is not a AnnotationLayer."
                )
        else:
            layer = AnnotationLayer(
                name=name,
                resolution=data_resolution,
                tags=tags,
                linked_segmentation=linked_segmentation,
                shader=shader,
                color=color,
                swap_visible_segments_on_move=swap_visible_segments_on_move,
            ).add_boxes(
                data,
                point_a_column=point_a_column,
                point_b_column=point_b_column,
                segment_column=segment_column,
                description_column=description_column,
                tag_column=tag_column,
                tag_bools=tag_bools,
                data_resolution=data_resolution,
            )
            self.add_layer(layer)
        return self

    def to_neuroglancer_state(self):
        if self.dimensions is None:
            if self.infer_coordinates and source_info.HAS_CLOUDVOLUME:
                self._dimensions = self._suggest_resolution_from_source()
            else:
                warnings.warn(
                    "No dimensions provided or inferred. Using a null CoordSpace which may cause unintended effects in Neuroglancer."
                )
                self._dimensions = CoordSpace()
        if not isinstance(self.dimensions, CoordSpace):
            self._dimensions = CoordSpace(self.dimensions)

        for layer in self.layers:
            if isinstance(layer, AnnotationLayer):
                if layer.resolution is None:
                    layer.resolution = self.dimensions.resolution

        if self.interactive:
            self._viewer = viewer.Viewer()
        else:
            self._viewer = UnservedViewer()
        if self.base_state:
            self._viewer.set_state(self.base_state)

        with self._viewer.txn() as s:
            if self.position:
                s.position = self.position
            elif self.infer_coordinates and s.position is None:
                s.position = self._suggest_position_from_source(
                    resolution=self.dimensions.resolution
                )
            if self.layout:
                s.layout = self.layout
            print(s.dimensions.rank)
            if s.dimensions.rank == 0:
                s.dimensions = self.dimensions.to_neuroglancer()
            if not self.base_state:
                s.cross_section_scale = self.scale_imagery
                s.projection_scale = self.scale_3d
                s.show_slices = self.show_slices
            for layer in self.layers:
                layer.apply_to_neuroglancer(s)

        return self._viewer

    def map(self, datamap: dict, inplace: bool = False) -> Self:
        """Apply a datamap to the viewer state to freeze the state of the viewer.
        Must be used if any layers use a datamap.
        By default, this will return a new viewer state with the datamap applied without changing the current object.

        Parameters
        ----------
        datamap : dict
            A dictionary mapping layer names to their corresponding datamaps.
        inplace : bool, optional
            If True, apply the datamap in place and return the modified viewer state.
            If False, return a new viewer state with the datamap applied.

        Returns
        -------
        Self
            A new ViewerState object with the datamap applied.
        """
        self._reset_viewer()
        if inplace:
            self._apply_datamaps(datamap)
            return self
        with self.with_datamap(datamap) as viewer_copy:
            return viewer_copy

    def to_dict(self) -> dict:
        """Return a dictionary representation of the viewer state.

        Returns
        -------
        dict
            A dictionary representation of the viewer state.
        """
        return self.viewer.state.to_json()

    def to_json_string(self, indent: int = 2) -> str:
        """Return a JSON string representation of the viewer state.

        Parameters
        ----------
        indent : int
            The number of spaces to use for indentation in the JSON string.
            Default is 2.

        Returns
        -------
        str
            A JSON string representation of the viewer state.
        """

        return json.dumps(self.to_dict(), indent=indent)

    def to_url(
        self,
        target_url: str = None,
        target_site: str = None,
        shorten: Union[bool, Literal["if_long"]] = False,
        client: Optional["caveclient.CAVEclient"] = None,
    ):
        """Return a URL representation of the viewer state.

        Parameters
        ----------
        target_url : str
            The base URL to use for the Neuroglancer state. If not provided,
            the default server URL will be used.
        target_site : str, optional
            The target site for the URL, based on the keys in site_utils.NEUROGLANCER . If not provided, the default server URL will be used.
        shorten: Union[bool, Literal["if_long"]], optional
            If True, the URL will be shortened using the CAVE link shortener service.
            If "if_long", the URL will only be shortened if it exceeds a certain length.
        client : Optional[caveclient.CAVEclient], optional
            The CAVE client to use for shortening the URL. If not provided, the URL will not be shortened.

        Returns
        -------
        str
            A URL representation of the viewer state.
        """
        if target_url is None:
            if target_site is None:
                if self.interactive:
                    return self.viewer.get_viewer_url()
                else:
                    target_url = self._target_url
            else:
                target_url = neuroglancer_url(target_site=target_site)

        url = neuroglancer.to_url(
            self.viewer.state,
            prefix=target_url,
        )
        if shorten == "if_long":
            if len(url) > MAX_URL_LENGTH:
                shorten = True
            else:
                shorten = False
        if shorten:
            url = self.to_link_shortener(client, target_url=target_url)
        return url

    def to_link(
        self,
        target_url: str = None,
        target_site: str = None,
        link_text: str = "Neuroglancer Link",
        shorten: Union[bool, Literal["if_long"]] = False,
        client: Optional["caveclient.CAVEclient"] = None,
    ):
        """Return an HTML link representation of the viewer state.

        Parameters
        ----------
        target_url : str
            The base URL to use for the Neuroglancer state. If not provided,
            the default server URL will be used.
        target_site : str, optional
            The target site for the URL, based on the keys in site_utils.NEUROGLANCER_SITES.
            If not provided, the default server URL will be used.
        link_text : str
            The text to display for the HTML link.
        shorten: Union[bool, Literal["if_long"]], optional
            If True, the URL will be shortened using the CAVE link shortener service.
            If "if_long", the URL will only be shortened if it exceeds a certain length.
        client : Optional[caveclient.CAVEclient], optional
            The CAVE client to use for shortening the URL. If not provided, the URL will not be shortened.

        Returns
        -------
        HTML.HTML
            An HTML link representation of the viewer state.
        """
        url = self.to_url(
            target_url=target_url,
            target_site=target_site,
            shorten=shorten,
            client=client,
        )
        return HTML(f'<a href="{url}" target="_blank">{link_text}</a>')

    def to_link_shortener(
        self,
        client: "caveclient.CAVEclient" = None,
        target_url: str = None,
        target_site: str = None,
    ) -> str:
        """Shorten the URL using the CAVE link shortener service.

        Parameters
        ----------
        client : CAVEclient
            The CAVE client to use for shortening the URL.
        target_url : str, optional
            The base URL to use for the Neuroglancer state. If not provided,
            the default server URL will be used.
        target_site : str, optional
            The target site for the URL, based on the keys in site_utils.NEUROGLANCER_SITES.
            If not provided, the default server URL will be used.

        Returns
        -------
        str
            A shortened URL representation of the viewer state.
        """
        if client is None:
            client = self._client
        if client is None:
            raise ValueError("A CAVEclient instance is required to shorten the URL.")

        state_id = client.state.upload_state_json(self.to_dict())
        if target_url is None:
            if target_site is None:
                if self.interactive:
                    return self.viewer.get_viewer_url()
                else:
                    target_url = self._target_url
            else:
                target_url = neuroglancer_url(target_site=target_site)
        return client.state.build_neuroglancer_url(state_id, target_url)

    def to_clipboard(
        self,
        target_url: str = None,
        target_site: str = None,
        shorten: Union[bool, Literal["if_long"]] = False,
        client: Optional["caveclient.CAVEclient"] = None,
    ) -> str:
        """Copy the viewer state URL to the system clipboard.

        Parameters
        ----------
        target_url : str
            The base URL to use for the Neuroglancer state. If not provided,
            the default server URL will be used.
        target_site : str, optional
            The target site for the URL, based on the keys in site_utils.NEUROGLANCER_SITES.
            If not provided, the default server URL will be used.
        shorten: Union[bool, Literal["if_long"]], optional
            If True, the URL will be shortened using the CAVE link shortener service.
            If "if_long", the URL will only be shortened if it exceeds a certain length.
        client : Optional[caveclient.CAVEclient], optional
            The CAVE client to use for shortening the URL. If not provided, the URL will not be shortened.

        Returns
        -------
        str
            The URL representation of the viewer state that has also been copied to the
            clipboard.
        """
        url = self.to_url(
            target_url=target_url,
            target_site=target_site,
            shorten=shorten,
            client=client,
        )
        pyperclip.copy(url)
        return url

    def to_browser(
        self,
        target_url: str = None,
        target_site: str = None,
        shorten: Union[bool, Literal["if_long"]] = False,
        client: Optional["caveclient.CAVEclient"] = None,
        new: int = 2,
        autoraise: bool = True,
        browser: Optional[str] = None,
    ):
        """Open the viewer state URL in a web browser.

        Parameters
        ----------
        target_url : str
            The base URL to use for the Neuroglancer state. If not provided,
            the default server URL will be used.
        target_site : str, optional
            The target site for the URL, based on the keys in site_utils.NEUROGLANCER_SITES.
            If not provided, the default server URL will be used.
        shorten: Union[bool, Literal["if_long"]], optional
            If True, the URL will be shortened using the CAVE link shortener service.
            If "if_long", the URL will only be shortened if it exceeds a certain length.
        client : Optional[caveclient.CAVEclient], optional
            The CAVE client to use for shortening the URL. If not provided, the URL will not be shortened.
        new : int, optional
            If new is 0, the url is opened in the same browser window if possible. If
            new is 1, a new browser window is opened if possible. If new is 2, a new
            browser page (“tab”) is opened if possible. Note that not all browsers
            support all values of new, and some may ignore this parameter.
        autoraise : bool, optional
            If True, the browser window will be raised to the front when opened. Note
            that under many window managers this will occur regardless of the setting
            of this variable.
        browser : Optional[str], optional
            The name of the browser to use. If None, the system default browser will be
            used. Note that the browser name needs to be registered on your system,
            see [webbrowser.register][] for more details.

        See Also
        --------
        [webbrowser.open][] : Opens a URL in the web browser.

        Returns
        -------
        str
            The URL representation of the viewer state that has also been opened in the
            browser.
        """
        url = self.to_url(
            target_url=target_url,
            target_site=target_site,
            shorten=shorten,
            client=client,
        )
        if browser is None:
            webbrowser.open(url, new=new, autoraise=autoraise)
        else:
            browser_controller = webbrowser.get(browser)
            browser_controller.open(url, new=new, autoraise=autoraise)
        return url



from nglui.statebuilder import site_utils


def test_site_utils():
    assert site_utils.neuroglancer_url() == site_utils.NEUROGLANCER_SITES.get(
        site_utils.DEFAULT_TARGET_SITE
    )
    assert site_utils.neuroglancer_url(
        target_site="spelunker"
    ) == site_utils.NEUROGLANCER_SITES.get("spelunker")
    assert (
        site_utils.neuroglancer_url(url="https://example.com") == "https://example.com"
    )

    site_utils.add_neuroglancer_site("testsite", "https://testsite.example.com")
    assert (
        site_utils.neuroglancer_url(target_site="testsite")
        == "https://testsite.example.com"
    )



import requests

ngl_info_endpoint = "{neuroglancer_endpoint}/version.json"


def get_ngl_info(ngl_url):
    """
    Get the version of neuroglancer running at a given endpoint.
    """
    try:
        r = requests.get(ngl_info_endpoint.format(neuroglancer_endpoint=ngl_url))
        r.raise_for_status()
        return r.json()
    except Exception as e:
        print(f"Error getting neuroglancer version: {e}")
        return None



import os
from typing import Optional

NEUROGLANCER_SITES = {
    "spelunker": "https://spelunker.cave-explorer.org/",
    "google": "https://neuroglancer-demo.appspot.com/",
}

DEFAULT_KEY = "DEFAULT_"
if "NGLUI_TARGET_SITE" in os.environ and "NGLUI_TARGET_URL" in os.environ:
    NEUROGLANCER_SITES[os.environ["NGLUI_TARGET_SITE"]] = os.environ["NGLUI_TARGET_URL"]
    DEFAULT_TARGET_SITE = os.environ["NGLUI_TARGET_SITE"]
else:
    DEFAULT_TARGET_SITE = "spelunker"
DEFAULT_TARGET_URL = NEUROGLANCER_SITES[DEFAULT_TARGET_SITE]
NEUROGLANCER_SITES[DEFAULT_KEY] = NEUROGLANCER_SITES[DEFAULT_TARGET_SITE]

MAX_URL_LENGTH = 1_750_000


def add_neuroglancer_site(
    site_name: str,
    site_url: str,
    set_default: bool = False,
) -> None:
    """
    Add a neuroglancer site to the list of available sites.

    Parameters
    ----------
    site_name : str
        Name of the neuroglancer site.
    site_url : str
        URL of the neuroglancer site.
    """
    if site_name in NEUROGLANCER_SITES:
        raise ValueError(f"Neuroglancer site {site_name} already exists")
    NEUROGLANCER_SITES[site_name] = site_url

    if set_default:
        set_default_neuroglancer_site(site_name)


def set_default_neuroglancer_site(
    site_name: str,
) -> None:
    """
    Set the default neuroglancer site.

    Parameters
    ----------
    site_name : str
        Name of the neuroglancer site to set as default.
    """
    if site_name not in NEUROGLANCER_SITES:
        raise ValueError(f"Neuroglancer site {site_name} does not exist")
    NEUROGLANCER_SITES[DEFAULT_KEY] = NEUROGLANCER_SITES[site_name]


def get_default_neuroglancer_site() -> str:
    """
    Get the default neuroglancer site URL.

    Returns
    -------
    dict
        Single element dictionary with the site and URL of the default neuroglancer site.
    """
    for key, value in NEUROGLANCER_SITES.items():
        if key == DEFAULT_KEY:
            continue
        if value == NEUROGLANCER_SITES[DEFAULT_KEY]:
            break
    return {key: NEUROGLANCER_SITES[DEFAULT_KEY]}


def neuroglancer_url(
    url: Optional[str] = None,
    target_site: Optional[str] = None,
) -> str:
    """
    Check neuroglancer info to determine which kind of site a neuroglancer URL is.
    If either url or target_site are provided, it will use these values, looking up target site
    from the fallback values in the config. Otherwise, it falls back to the value of
    "target_url" in the config.

    Parameters
    ----------
    url : str, optional
        URL to check, by default None
    target_site : str, optional
        Target site to check, by default None

    Returns
    -------
    str
        URL of the neuroglancer viewer
    """
    if url is not None:
        return url
    if target_site is None:
        target_site = DEFAULT_KEY
    if target_site in NEUROGLANCER_SITES:
        return NEUROGLANCER_SITES[target_site]
    else:
        raise ValueError(
            f"Neuroglancer site {target_site} not found. Available sites: {list(NEUROGLANCER_SITES.keys()).remove(DEFAULT_KEY)}"
        )


def get_neuroglancer_sites() -> dict:
    """
    Get the list of available neuroglancer sites.

    Returns
    -------
    dict
        List of available neuroglancer URLs and their names.
    """
    ngl_sites = NEUROGLANCER_SITES.copy()
    ngl_sites.pop(DEFAULT_KEY, None)
    return ngl_sites



import pytest
from pytest_mock import mocker

from nglui.statebuilder import *


def test_basic_viewer(client_full, mocker):
    viewer = (
        ViewerState(dimensions=[4, 4, 40])
        .add_layer(
            ImageLayer(
                source=client_full.info.image_source(),
                name="TestImageLayer",
            )
        )
        .add_layer(
            SegmentationLayer(
                source=client_full.info.segmentation_source(),
                name="TestSegmentationLayer",
                pick=False,
            )
        )
        .add_segments(name="TestSegmentationLayer", segments=[1, 2, 3])
        .add_layer(AnnotationLayer(name="some_annotations"))
        .add_annotation_layer(
            name="other_annotations",
            linked_segmentation="TestSegmentationLayer",
        )
        .set_viewer_properties(
            position=[1, 1, 1],
            scale_imagery=10,
            scale_3d=1000.4,
            selected_layer="some_annotations",
            layout="xy",
        )
        .add_layer(
            AnnotationLayer(
                name="TestAnnotationLayer",
                source="precomputed://gs://pathtoannotations",
            )
        )
    )
    assert viewer.layers[0].name == "TestImageLayer"
    assert viewer.layers["TestImageLayer"].source == client_full.info.image_source()
    assert isinstance(viewer.to_dict(), dict)
    assert viewer.to_json_string() is not None
    assert "https" in viewer.to_url()
    viewer.to_link()

    mocker.patch.object(
        client_full.state,
        "upload_state_json",
        return_value=12345,
    )

    assert "12345" in viewer.to_link_shortener(client=client_full)

    site_utils.add_neuroglancer_site(
        site_name="test_site",
        site_url="https://test.neuroglancer.com",
    )
    assert "https://test.neuroglancer.com" in viewer.to_url(target_site="test_site")


def test_adding_points(client_full, soma_df):
    viewer = (
        ViewerState(infer_coordinates=False)
        .add_layers_from_client(
            client_full,
        )
        .add_segments_from_data(
            data=soma_df.head(2),
            segment_column="pt_root_id",
        )
        .add_points(
            data=soma_df,
            name="soma_points",
            point_column="pt_position",
            segment_column="pt_root_id",
            tag_column="cell_type",
        )
    )
    assert len(viewer.layers["segmentation"].segments) == 2
    state = viewer.to_dict()
    assert len(state["layers"][2]["annotations"]) == len(soma_df)


def test_adding_lines(client_full, pre_syn_df):
    viewer = (
        ViewerState(infer_coordinates=False)
        .add_layers_from_client(
            client_full,
        )
        .add_lines(
            data=pre_syn_df,
            name="pre_syn_lines",
            point_a_column="pre_pt_position",
            point_b_column="post_pt_position",
            segment_column="post_pt_root_id",
        )
    )
    state = viewer.to_dict()
    assert len(state["layers"][2]["annotations"]) == len(pre_syn_df)


def test_translated_source(img_path):
    img35_translate = Source(
        url=img_path,
        transform=CoordSpaceTransform(
            output_dimensions=[4, 4, 40],
            matrix=[
                [1, 0, 0.5, 1000],
                [0, 1, 0, 2000.0],
                [0, 0, 1, -1000.0],
            ],
        ),
    )

    img_layer_translated = ImageLayer("imagery_translated", img35_translate)

    img_layer_translated.to_dict()



from __future__ import annotations

import copy
import json
from abc import ABC, abstractmethod
from contextlib import contextmanager
from functools import partial
from typing import Optional, Union

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

import caveclient
import numpy as np
import pandas as pd
from attrs import define, field
from neuroglancer import Viewer, viewer_state
from neuroglancer.coordinate_space import CoordinateSpace
from neuroglancer.json_wrappers import optional, wrapped_property

from ..segmentprops import SegmentProperties
from .ngl_annotations import (
    MAX_TAG_COUNT,
    AnnotationBase,
    BoundingBoxAnnotation,
    EllipsoidAnnotation,
    LineAnnotation,
    PointAnnotation,
    make_annotation_properties,
    make_bindings,
    strip_numpy_types,
)
from .shaders import DEFAULT_SHADER_MAP
from .utils import (
    is_dict_like,
    is_list_like,
    parse_color,
    parse_graphene_header,
    parse_graphene_image_url,
    split_point_columns,
)

# Monkey-patch AnnotationLayer to add swap_visible_segments_on_move property
viewer_state.AnnotationLayer.swap_visible_segments_on_move = (
    viewer_state.AnnotationLayer.swapVisibleSegmentsOnMove
) = wrapped_property("swapVisbleSegmentsOnMove", optional(bool, True))


class UnmappedDataError(Exception):
    """Exception raised when a layer is not fully mapped to a datamap."""


@define
class DataMap:
    """Class to defer mapping of dataframes to a ViewerState or Layer.

    Parameters
    ----------
    key : str, optional
        The key to use for the datamap, which is used to label data when using the `map` or `with_map` classes.
        If no key is provided, a default key of "None" is used and assumed if no dictionary is provided later.
    """

    key = field(default=None, type=Optional[str])
    priority = field(default=10, type=int, init=False, repr=False)

    def _adjust_priority(self, value: int) -> Self:
        """Make the datamap high priority."""
        self.priority = value
        return self


@define
class CoordSpace:
    """Coordinate space for Neuroglancer.
    Parameters
    ----------
    resolution : list[int], optional
        The resolution of the coordinate space. Default is None, which will raise an error if not set.
    units : str or list[str], optional
        The units of the coordinate space. Default is "nm". If a single string is provided, it will be repeated for each dimension.
    names : list[str], optional
        The names of the dimensions. Default is ["x", "y", "z"]. If the length of names does not match the length of resolution, an error will be raised.
    """

    resolution = field(default=None, type=list[int])
    units = field(default="nm", type=str)
    names = field(factory=lambda: ["x", "y", "z"], type=list[str])

    def __attrs_post_init__(self):
        if self.resolution is None:
            self.resolution = []
            self.units = []
            self.names = []
        if isinstance(self.units, str):
            self.units = [self.units] * len(self.resolution)
            if len(self.names) != len(self.units):
                raise ValueError("Length of names and unit must match")
            if len(self.resolution) != len(self.names):
                raise ValueError("Length of resolution and names must match")

    def to_neuroglancer(self) -> CoordinateSpace:
        if self.resolution is None:
            raise ValueError("Resolution must be set before converting to Neuroglancer")
        return CoordinateSpace(
            units=self.units, scales=self.resolution, names=self.names
        )


@define
class CoordSpaceTransform:
    """Coordinate space transform for Neuroglancer.

    Parameters
    ----------
    output_dimensions : list[int] or CoordSpace, optional
        The output dimensions of the transform. If a list, it will be converted to a CoordSpace.
        Default is None, which will create a transform with no output dimensions.
    input_dimensions : list[int] or CoordSpace, optional
        The input dimensions of the transform. If a list, it will be converted to a CoordSpace.
        Default is None, which will create a transform with no input dimensions.
    matrix : list[list[float]], optional
        The transformation matrix. Default is None, which will create an identity matrix.
        The matrix should be a 4x4 list of lists, where the last row is [0, 0, 0, 1].
        If None, an identity matrix will be created based on the output dimensions.
    """

    output_dimensions = field(default=None, type=list[int])
    input_dimensions = field(default=None, type=list[int])
    matrix = field(default=None, type=list[list[float]])

    def __attrs_post_init__(self):
        if self.output_dimensions is None:
            self.output_dimensions = None
        elif not isinstance(self.output_dimensions, CoordSpace):
            self.output_dimensions = CoordSpace(resolution=self.output_dimensions)

    def to_neuroglancer(self):
        if self.output_dimensions is None:
            return None
        if self.input_dimensions is not None:
            input_dims = self.input_dimensions.to_neuroglancer()
        else:
            input_dims = None
        return viewer_state.CoordinateSpaceTransform(
            output_dimensions=self.output_dimensions.to_neuroglancer(),
            input_dimensions=input_dims,
            matrix=self.matrix,
        )


@define
class Source:
    """
    Configuration for a Neuroglancer data source.

    Parameters
    ----------
    url : str
        The URL of the data source.
    resolution : list, optional
        The resolution of the data source. Default is None.
    transform : CoordSpaceTransform, optional
        The coordinate space transform for the data source. Default is None, which will create a transform based on the resolution.
    subsources : dict, optional
        A dictionary of subsources for the data source (e.g. meshes, skeletons, etc). Default is None, which includes all subsources.
    enable_default_subsources : bool, optional
        Whether to enable default subsources. Default is True, which includes all default subsources.
    """

    url = field(type=str)
    resolution = field(default=None, type=list)
    transform = field(default=None, type=CoordSpaceTransform)
    subsources = field(default=None, type=dict)
    enable_default_subsources = field(default=True, type=bool)

    def to_neuroglancer(self) -> dict:
        """Convert the Source to a neuroglancer-python object."""
        if self.transform is None:
            self.transform = CoordSpaceTransform(output_dimensions=self.resolution)
        return viewer_state.LayerDataSource(
            url=self.url,
            transform=self.transform.to_neuroglancer(),
            subsources=self.subsources,
            enable_default_subsources=self.enable_default_subsources,
        )


@define
class Layer(ABC):
    name = field(type=str)
    resolution = field(default=None, type=list, kw_only=True, repr=False)
    visible = field(default=True, type=bool, kw_only=True, repr=False)
    archived = field(default=False, type=bool, kw_only=True, repr=False)
    pick = field(default=True, type=bool, kw_only=True, repr=False)
    _datamaps = field(factory=dict, type=dict, init=False, repr=False)
    _datamap_priority = field(factory=dict, type=dict, init=False, repr=False)

    @contextmanager
    def with_datamap(self, datamap: dict):
        if datamap is None:
            yield self
        else:
            if not isinstance(datamap, dict):
                datamap = {None: datamap}

            layer_copy = copy.deepcopy(self)
            dm_keys = list(layer_copy._datamap_priority.keys())
            dm_priorities = list(layer_copy._datamap_priority.values())
            sorted_keys = [
                x for x, _ in sorted(zip(dm_keys, dm_priorities), key=lambda x: x[1])
            ]

            for k in sorted_keys:
                if k in datamap:
                    func = layer_copy._datamaps.pop(k)
                    func(datamap.pop(k))
            yield layer_copy

    @property
    def is_static(self) -> bool:
        """Check if the layer is static."""
        return len(self._datamaps) == 0

    def _check_fully_mapped(self):
        """Check if the layer is fully mapped."""
        if not self.is_static:
            raise UnmappedDataError(
                f"Layer '{self.name}' has datamaps registered but no datamap provided: {list(self._datamaps.keys())}"
            )

    def _register_datamap(self, key, func, **kwargs):
        """Register a function to be called when the datamap is set.

        Parameters
        ----------
        key: DataMap
            Object with a key and priority for the datamap.
        func : Callable
            The function to register.
        kwargs:
            Additional keyword arguments to pass to the function.
        """
        self._datamaps[key.key] = partial(func, **kwargs)
        self._datamap_priority[key.key] = key.priority

    def _apply_datamaps(self, datamap: dict):
        """Map"""
        for k, v in datamap.items():
            if k in self._datamaps:
                self._datamaps.pop(k)(v)  # Apply the function with the remaining values

    def map(self, datamap: dict, inplace: bool = False) -> Self:
        """Map the layer to a datamap.

        Parameters
        ----------
        datamap : dict
            The datamap to map the layer to.
        inplace: bool, optional
            Whether to modify the layer in place or return a new layer. Default is False.
        """
        if inplace:
            self._apply_datamaps(datamap)
            return self
        else:
            with self.with_datamap(datamap) as layer:
                return layer

    @abstractmethod
    def to_neuroglancer_layer(self):
        self._check_fully_mapped()

    def to_dict(self, with_name: bool = True) -> dict:
        """Convert the layer to a dictionary.
        Parameters
        ----------
        with_name : bool, optional
            Whether to include the name, visibility, and archived states of the layer in the dictionary, by default True.
            These are not typically included until the layer is part of a state, but adding them in allows the resulting data to be passed to a viewer state string.
        datamap: dict, optional
            A dictionary with keys being names of DataMap parameters and values being the data to be passed.
            Must be provided if the layer has any datamaps registered.

        Returns
        -------
        dict
            The layer as a dictionary.
        """
        layer_dict = self.to_neuroglancer_layer().to_json()
        if with_name:
            layer_dict["name"] = self.name
            layer_dict["visible"] = self.visible
            layer_dict["archived"] = self.archived
        return layer_dict

    def to_json(self, with_name: bool = True, indent: int = 2):
        return json.dumps(self.to_dict(with_name=with_name), indent=indent)

    def _apply_to_neuroglancer_state(self, s):
        if self.name in s.layers:
            raise ValueError(
                f"Layer {self.name} already exists in the viewer. Please use a different name."
            )
        s.layers[self.name] = self.to_neuroglancer_layer()
        ll = s.layers[self.name]
        ll.visible = self.visible
        ll.archived = self.archived

    def _apply_to_neuroglancer(self, viewer):
        # Opens context or not depending on if the object is a viewer or a (presumed within-context) state
        self._check_fully_mapped()
        if isinstance(viewer, Viewer):
            with viewer.txn() as s:
                self._apply_to_neuroglancer_state(s)
        elif isinstance(viewer, viewer_state.ViewerState):
            self._apply_to_neuroglancer_state(viewer)


@define
class LayerWithSource(Layer):
    source = field(factory=list, type=Union[list, Source])

    def __attrs_post_init__(self):
        if isinstance(self.source, DataMap):
            self.source._adjust_priority(1)
            self._register_datamap(
                key=self.source,
                func=self.add_source,
                resolution=self.resolution,
            )
            self.source = []

    def add_source(
        self,
        source: Union[str, list, Source, DataMap],
        resolution: Optional[list] = None,
    ) -> Self:
        """Add a source to the layer.

        Parameters
        ----------
        source : str or Source
            The source to add.
        resolution : list or np.typing.ArrayLike, optional
            The resolution of the source. Default is None.

        """
        if self.source is None:
            self.source = []
        if isinstance(source, DataMap):
            if len(self.source) > 0:
                raise ValueError(
                    "Cannot add a DataMap source to a layer that already has sources."
                )
            self._register_datamap(
                key=source._adjust_priority(1),
                func=self.add_source,
                resolution=resolution,
            )
            return self  # Return early for DataMap
        if isinstance(self.source, Source) or isinstance(self.source, str):
            self.source = [self.source]
        if isinstance(source, str):
            self.source.append(Source(url=source, resolution=resolution))
        elif isinstance(source, Source):
            self.source.append(source)
        elif is_list_like(source):
            for src in source:
                if isinstance(src, str):
                    self.source.append(Source(url=src, resolution=resolution))
                elif isinstance(src, Source):
                    self.source.append(src)
                else:
                    raise ValueError(
                        "Invalid source type. Must be str, list, Source, or DataMap."
                    )
        else:
            raise ValueError(
                "Invalid source type. Must be str, list, Source, or DataMap."
            )
        return self


def _handle_source(source, resolution=None, image_layer=False):
    "Convert one or more sources to a Source object."
    if image_layer:
        auth_parse_func = parse_graphene_image_url
    else:
        auth_parse_func = parse_graphene_header
    if isinstance(source, str):
        return Source(url=auth_parse_func(source), resolution=resolution)
    elif isinstance(source, Source):
        source.url = auth_parse_func(source.url)
        return source
    elif is_list_like(source):
        return [
            _handle_source(src, resolution, image_layer=image_layer) for src in source
        ]
    else:
        raise ValueError("Invalid source type. Must be str or Source.")


def _handle_annotations(annos, tag_map=None, resolution=None):
    "Convert a multi-url source to a Source object."
    if annos is None:
        return []
    elif len(annos) == 0:
        return []
    return [
        anno.to_neuroglancer(tag_map=tag_map, layer_resolution=resolution)
        if issubclass(type(anno), AnnotationBase)
        else anno
        for anno in annos
    ]


def source_to_neuroglancer(source, resolution=None, image_layer=False):
    "Convert a possibly multi-url source to a Neuroglancer-compatible format."

    source = _handle_source(source, resolution=resolution, image_layer=image_layer)
    if isinstance(source, list):
        return [src.to_neuroglancer() for src in source]
    else:
        return source.to_neuroglancer()


def segments_to_neuroglancer(segments):
    "Convert a flat or mixed-visibility segment list to Neuroglancer."
    if segments is None:
        return viewer_state.StarredSegments()
    if is_list_like(segments) or is_dict_like(segments):
        # Annoying processing to avoid np.True_/np.False_ types
        return viewer_state.StarredSegments(strip_numpy_types(segments))
    else:
        return segments


def _handle_linked_segmentation(segmentation_layer) -> dict:
    if segmentation_layer is None:
        return None
    elif isinstance(segmentation_layer, dict):
        return segmentation_layer
    elif isinstance(segmentation_layer, SegmentationLayer):
        return {"segments": segmentation_layer.name}
    elif isinstance(segmentation_layer, str):
        return {"segments": segmentation_layer}
    else:
        raise ValueError(
            "Invalid linked segmentation layer type. Must be str or SegmentationLayer."
        )


def _handle_filter_by_segmentation(
    filter: Optional[Union[list, bool, str]],
    linked_segmentation: Optional[Union[list, bool, str]],
):
    linked_seg = _handle_linked_segmentation(linked_segmentation)
    if filter is True:
        return [x for x in linked_seg.keys()]
    elif filter is False:
        return []
    elif is_list_like(filter):
        return linked_seg


@define
class RawLayer(Layer):
    """Configuration for a Neuroglancer raw layer from a data source"""

    name = field(type=str, kw_only=True, default=None)
    json_data = field(type=dict)

    def __attrs_post_init__(self):
        self.json_data = copy.deepcopy(self.json_data)
        if self.name is None and "name" in self.json_data:
            self.name = self.json_data.pop("name")
        if self.visible is None and "visible" in self.json_data:
            self.visible = self.json_data.pop("visible")
        if self.archived is None and "archived" in self.json_data:
            self.archived = self.json_data.pop("archived")

    @property
    def is_static(self):
        return True

    def to_neuroglancer_layer(self):
        return viewer_state.Layer(json_data=self.json_data)

    def apply_to_neuroglancer(self, viewer):
        return self._apply_to_neuroglancer(viewer)

    def remap_sources(self, source_map: dict) -> Self:
        """
        Remap the sources in the layer's JSON data using the provided source map.

        Parameters
        ----------
        source_map : dict
            A mapping from old source paths to new source identifiers.

        Returns
        -------
        Self
            The updated layer with remapped sources.
        """
        if source_map:
            if isinstance(self.json_data.get("source"), str):
                self.json_data["source"] = source_map.get(
                    self.json_data["source"], self.json_data["source"]
                )
            elif isinstance(self.json_data.get("source"), list):
                for ii, src in enumerate(self.json_data["source"]):
                    if isinstance(src, str):
                        self.json_data["source"][ii] = source_map.get(src, src)
                    elif isinstance(src, dict):
                        self.json_data["source"][ii]["url"] = source_map.get(
                            self.json_data["source"][ii]["url"],
                            self.json_data["source"][ii]["url"],
                        )
        return self


@define
class ImageLayer(LayerWithSource):
    """Configuration for a Neuroglancer image layer.

    Parameters
    ----------
    name : str, optional
        The name of the layer. Default is "img".
    source : Source or list of Source, optional
        The source of the image data. Can be a Source object or a list of Source objects.
    shader : str, optional
        The shader to use for rendering the image. Default is None, which will use the default shader.
    color : list, optional
        The color to use for the image layer. Default is None, which will use the default color.
    opacity : float, optional
        The opacity of the image layer. Default is 1.0.
    blend : str, optional
        The blending mode for the image layer. Default is None, which will use the default blend mode.
    volume_rendering_mode : str, optional
        The volume rendering mode for the image layer. Default is None, which will use the default mode.
    volume_rendering_gain : float, optional
        The gain for volume rendering. Default is None, which will use the default gain.
    volume_rendering_depth_samples : int, optional
        The number of depth samples for volume rendering. Default is None, which will use the default number of samples.
    cross_section_render_scale : float, optional
        The scale for cross-section rendering. Default is None, which will use the default scale.
    pick: bool, optional
        Whether to allow cursor interaction with meshes and skeletons. Default is True.
    """

    name = field(default="img", type=str)
    source = field(factory=list, type=Union[list, Source])
    shader = field(default=None, type=Optional[str], kw_only=True, repr=False)
    color = field(default=None, type=list, kw_only=True, repr=False)
    opacity = field(default=1.0, type=float, kw_only=True, repr=False)
    blend = field(default=None, type=str, kw_only=True, repr=False)
    volume_rendering_mode = field(default=None, type=str, kw_only=True, repr=False)
    volume_rendering_gain = field(default=None, type=float, kw_only=True, repr=False)
    volume_rendering_depth_samples = field(
        default=None, type=int, kw_only=True, repr=False
    )
    cross_section_render_scale = field(
        default=None, type=float, kw_only=True, repr=False
    )

    def __attrs_post_init__(self):
        self.color = parse_color(self.color)
        super().__attrs_post_init__()

    def to_neuroglancer_layer(self) -> viewer_state.ImageLayer:
        super().to_neuroglancer_layer()
        if self.shader is None:
            return viewer_state.ImageLayer(
                source=source_to_neuroglancer(
                    self.source, resolution=self.resolution, image_layer=True
                ),
                annotation_color=self.color,
            )
        else:
            return viewer_state.ImageLayer(
                source=source_to_neuroglancer(
                    self.source, resolution=self.resolution, image_layer=True
                ),
                shader=self.shader,
                annotation_color=self.color,
            )

    def apply_to_neuroglancer(self, viewer: Viewer) -> viewer_state.ImageLayer:
        "Can be a viewer or a viewer.txn()-context state"
        self._apply_to_neuroglancer(viewer)

    def add_from_client(
        self,
        client: caveclient.CAVEclient,
    ) -> Self:
        """Add an image layer source from caveclient.

        Parameters
        ----------
        client : CAVEclient
            The CAVEclient object.

        Returns
        -------
        ImageLayer
            An ImageLayer object.

        """
        return self.add_source(
            source=client.info.image_source(),
            resolution=client.info.viewer_resolution(),
        )

    def add_shader(self, shader: str) -> Self:
        """Add a shader to the layer.

            Parameters
            ----------
            shader : str
                The shader to add.

        #"""
        self.shader = shader
        return self


@define
class SegmentationLayer(LayerWithSource):
    """Configuration for a Neuroglancer segmentation layer.

    Parameters
    ----------
    name : str, optional
        The name of the layer. Default is "seg".
    source : Source or str, optional
        The source of the segmentation data. Can be a Source object or a string URL.
    segments : list or dict or VisibleSegments, optional
        The segments to display in the layer. Can be a list of segment IDs, a dictionary with segment IDs as keys and visibility as values, or a VisibleSegments object.
    color : list, optional
        The color to use for the segments. Default is None, which will use the default color.
    hide_segment_zero : bool, optional
        Whether to hide segment zero, which is typically treated as "no segmentation". Default is True.
    selected_alpha : float, optional
        The transparency value for selected segments in the 2d views. Default is 0.2.
    not_selected_alpha : float, optional
        The transparency value for unselected segments in the 2d views. Default is 0.0.
    alpha_3d : float, optional
        The transparency value for segments in the 3d view. Default is 0.9.
    mesh_silhouette : float, optional
        The silhouette rendering value for the mesh. Default is 0.0.
    segment_colors : dict, optional
        A dictionary mapping segment IDs to colors. Default is None, which will use the default colors.
    shader : str, optional
        The shader to use for rendering the skeletons if a skeleton source is provided. Default is None, which will use the default shader.
    pick: bool, optional
        Whether to allow cursor interaction with meshes and skeletons. Default is True.

    """

    name = field(default="seg", type=str)
    source = field(default=None, type=Union[str, Source])
    segments = field(
        factory=list,
        type=Union[list, dict, viewer_state.VisibleSegments],
        kw_only=True,
        repr=False,
    )
    color = field(default=None, type=list, kw_only=True, repr=False)
    hide_segment_zero = field(default=True, type=bool, kw_only=True, repr=False)
    selected_alpha = field(default=0.2, type=float, kw_only=True, repr=False)
    not_selected_alpha = field(default=0.0, type=float, kw_only=True, repr=False)
    alpha_3d = field(default=0.9, type=float, kw_only=True, repr=False)
    mesh_silhouette = field(default=0.0, type=float, kw_only=True, repr=False)
    segment_colors = field(default=None, type=dict, kw_only=True, repr=False)
    shader = field(default=None, type=str, kw_only=True, repr=False)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if isinstance(self.segments, DataMap):
            self._register_datamap(
                key=self.segments,
                priority=self.segments.priority,
                func=self.add_segments,
            )
            self.segments = list()
        self.color = parse_color(self.color)

    def to_neuroglancer_layer(self) -> viewer_state.SegmentationLayer:
        super().to_neuroglancer_layer()
        if self.shader is None:
            return viewer_state.SegmentationLayer(
                source=source_to_neuroglancer(self.source, resolution=self.resolution),
                visible_segments=segments_to_neuroglancer(self.segments),
                annotation_color=self.color,
                selected_alpha=self.selected_alpha,
                not_selected_alpha=self.not_selected_alpha,
                object_alpha=self.alpha_3d,
                segment_colors=self.segment_colors,
                mesh_silhouette_rendering=self.mesh_silhouette,
                pick=self.pick,
            )
        else:
            return viewer_state.SegmentationLayer(
                source=source_to_neuroglancer(self.source, resolution=self.resolution),
                visible_segments=segments_to_neuroglancer(self.segments),
                annotation_color=self.color,
                selected_alpha=self.selected_alpha,
                not_selected_alpha=self.not_selected_alpha,
                object_alpha=self.alpha_3d,
                segment_colors=self.segment_colors,
                mesh_silhouette_rendering=self.mesh_silhouette,
                skeleton_shader=self.shader,
                pick=self.pick,
            )

    def apply_to_neuroglancer(self, viewer) -> viewer_state.SegmentationLayer:
        "Can be a viewer or a viewer.txn()-context state"
        return self._apply_to_neuroglancer(viewer)

    def add_from_client(
        self,
        client: caveclient.CAVEclient,
        add_skeleton_source: bool = True,
    ) -> Self:
        """Add a segmentation layer source from caveclient.

        Parameters
        ----------
        client : CAVEclient
            The CAVEclient object.

        Returns
        -------
        SegmentationLayer
            A SegmentationLayer object.

        """
        self.add_source(
            source=client.info.segmentation_source(),
            resolution=client.info.viewer_resolution(),
        )
        if (
            add_skeleton_source
            and client.info.get_datastack_info().get("skeleton_source") is not None
        ):
            self.add_source(
                source=client.info.get_datastack_info().get("skeleton_source")
            )
        return self

    def add_segments(
        self,
        segments: Union[list, dict, viewer_state.VisibleSegments],
        visible: Optional[list] = None,
    ) -> Self:
        """Add segment ids to the layer.

        Parameters
        ----------
        segments : list or dict or VisibleSegments
            The segments to add. If a dict, the keys are the segment IDs and the values are the boolean visibility.
        visible: list, optional
            The visibility of the segments, assumed to be True if not provided.
            Should be the same length as segments, and segments should be a list of the same length.

        Returns
        -------
        SegmentationLayer
            The SegmentationLayer object.
        """
        if isinstance(segments, DataMap):
            self._register_datamap(
                key=segments,
                func=self.add_segments,
                visible=visible,
            )
            return self
        old_segments = dict(segments_to_neuroglancer(self.segments))
        if visible is not None:
            segments = {s: v for s, v in zip(segments, visible)}
        elif not is_dict_like(segments):
            segments = {s: True for s in segments}
        new_segments = dict(segments_to_neuroglancer(segments))
        self.segments = old_segments | new_segments
        return self

    def add_segment_colors(
        self,
        segment_colors: dict,
    ) -> Self:
        """Add segment colors to the layer.

        Parameters
        ----------
        segment_colors : dict
            The segment colors to add.

        Returns
        -------
        SegmentationLayer
            The SegmentationLayer object with the added segment colors.
        """
        if self.segment_colors is None:
            self.segment_colors = {}
        parsed_colors = {k: parse_color(v) for k, v in segment_colors.items()}
        self.segment_colors.update(parsed_colors)
        return self

    def add_segments_from_data(
        self,
        data: Union[pd.DataFrame, DataMap],
        segment_column: str,
        visible_column: str = None,
        color_column: str = None,
    ) -> Self:
        """Create a SegmentationLayer from a DataFrame.

        Parameters
        ----------
        data : pd.DataFrame
            The DataFrame containing the segmentation data.
        name : str, optional
            The name of the layer. Default is "seg".
        color : str, optional
            The color to use.
        shader : str, optional
            The shader to use.

        Returns
        -------
        SegmentationLayer
            A SegmentationLayer object.
        """
        if isinstance(data, DataMap):
            self._register_datamap(
                key=data,
                func=self.segments_from_dataframe,
                segment_column=segment_column,
                visible_column=visible_column,
                color_column=color_column,
            )
            return self
        segments = data[segment_column].values
        if visible_column is not None:
            segments = {s: v for s, v in zip(segments, data[visible_column].values)}
        if color_column is not None:
            self.add_segment_colors(
                {s: parse_color(c) for s, c in zip(segments, data[color_column].values)}
            )

        self.add_segments(segments)
        return self

    def add_segment_properties(
        self,
        data: Union[pd.DataFrame, DataMap],
        client: "caveclient.CAVEclient",
        id_column: str = "pt_root_id",
        label_column: Optional[Union[str, list]] = None,
        description_column: Optional[str] = None,
        string_columns: Optional[list] = None,
        number_columns: Optional[list] = None,
        tag_value_columns: Optional[list] = None,
        tag_bool_columns: Optional[list] = None,
        tag_descriptions: Optional[list] = None,
        allow_disambiguation: bool = True,
        label_separator: str = "_",
        label_format_map: Optional[str] = None,
        prepend_col_name: bool = False,
        random_columns: Optional[int] = None,
        random_column_prefix: str = None,
        dry_run: bool = False,
    ) -> Self:
        """Upload segment properties and add to the layer.
        If you already have a segment properties cloud path, use `add_source` to add it to the layer.

        Parameters
        ----------
        data : pd.DataFrame
            The DataFrame containing the segment properties.
        client : CAVEclient
            The CAVEclient object needed to upload .
        id_column : str, optional
            The column name for the segment IDs. Default is 'pt_root_id'.
        label_column : str or list, optional
            The column name(s) for the segment labels. Default is None.
        description_column : str, optional
            The column name for the segment descriptions. Default is None.
        string_columns : list, optional
            The column names for string properties. Default is None.
        number_columns : list, optional
            The column names for number properties. Default is None.
        tag_value_columns : list, optional
            The column names for tag value properties. Default is None.
        tag_bool_columns : list, optional
            The column names for tag boolean properties. Default is None.
        tag_descriptions : list, optional
            The descriptions for the tags. Default is None.
        allow_disambiguation : bool, optional
            Whether to allow disambiguation of segment IDs. Default is True.
        label_separator : str, optional
            The separator for label formatting. Default is "_".
        label_format_map : str, optional
            The format map for labels. Default is None.
        prepend_col_name: bool, optional
            Whether to prepend the column name to the label. Default is False.
        random_columns: int, optional
            Number of random columns to add. Default is None.
        random_column_prefix: str, optional
            Name prefix of the random columns. Default is None.
        dry_run: bool, optional
            If dry run is true, build but do not actually upload and instead use a placeholder source text. Default is False.

        Returns
        -------
        SegmentationLayer
            A SegmentationLayer object with added segment properties.

        """
        if isinstance(data, DataMap):
            self._register_datamap(
                key=data,
                func=self.segment_properties,
                client=client,
                id_column=id_column,
                label_column=label_column,
                description_column=description_column,
                string_columns=string_columns,
                number_columns=number_columns,
                tag_value_columns=tag_value_columns,
                tag_bool_columns=tag_bool_columns,
                tag_descriptions=tag_descriptions,
                allow_disambiguation=allow_disambiguation,
                label_separator=label_separator,
                label_format_map=label_format_map,
                prepend_col_name=prepend_col_name,
                random_columns=random_columns,
                random_column_prefix=random_column_prefix,
            )
        segprops = SegmentProperties.from_dataframe(
            df=data,
            id_col=id_column,
            label_col=label_column,
            description_col=description_column,
            string_cols=string_columns,
            number_cols=number_columns,
            tag_value_cols=tag_value_columns,
            tag_bool_cols=tag_bool_columns,
            tag_descriptions=tag_descriptions,
            allow_disambiguation=allow_disambiguation,
            label_separator=label_separator,
            label_format_map=label_format_map,
            prepend_col_name=prepend_col_name,
            random_columns=random_columns,
            random_column_prefix=random_column_prefix,
        )
        if not dry_run:
            prop_id = client.state.upload_property_json(segprops.to_dict())
            prop_url = client.state.build_neuroglancer_url(
                prop_id, format_properties=True
            )
        else:
            prop_url = "DRYRUN_SEGMENT_PROPERTIES"
        self.add_source(prop_url)
        return self

    def set_view_options(
        self,
        selected_alpha: Optional[float] = None,
        not_selected_alpha: Optional[float] = None,
        alpha_3d: Optional[float] = None,
        mesh_silhouette: Optional[float] = None,
    ) -> Self:
        """Set the view options for the layer.

        Parameters
        ----------
        selected_alpha : float, optional
            The alpha value for selected segments. Default is 0.2.
        not_selected_alpha : float, optional
            The alpha value for not selected segments. Default is 0.0.
        alpha_3d : float, optional
            The alpha value for 3D segments. Default is 0.9.
        mesh_silhouette : float, optional
            The silhouette value for the mesh. Default is 0.0.

        Returns
        -------
        SegmentationLayer
            A SegmentationLayer object with updated view options.

        """
        if selected_alpha:
            self.selected_alpha = selected_alpha
        if not_selected_alpha:
            self.not_selected_alpha = not_selected_alpha
        if alpha_3d:
            self.alpha_3d = alpha_3d
        if mesh_silhouette:
            self.mesh_silhouette = mesh_silhouette
        return self

    def add_shader(self, shader: str) -> Self:
        """Add a shader to the layer.

        Parameters
        ----------
        shader : str
            The shader to add.

        """
        self.shader = shader
        return self

    def add_default_skeleton_shader(self) -> Self:
        """Add a default skeleton shader with desaturated axons to the layer."""
        return self.add_shader(DEFAULT_SHADER_MAP["skeleton_compartments"])


@define
class AnnotationLayer(LayerWithSource):
    name = field(default="anno", type=str)
    source = field(default=None, type=Union[str, Source])
    resolution = field(default=None, type=list, kw_only=True, repr=False)
    color = field(default=None, type=list, kw_only=True, repr=False)
    annotations = field(factory=list, type=list, kw_only=True, repr=False)
    shader = field(default=None, type=str, kw_only=True, repr=False)
    tags = field(factory=list, type=list, kw_only=True, repr=False)
    filter_by_segmentation = field(
        default=False, type=Union[str, bool, list], kw_only=True, repr=False
    )
    linked_segmentation = field(
        default=None, type=Union[str, dict], kw_only=True, repr=False
    )
    set_position = field(default=True, type=bool, kw_only=True, repr=False)
    swap_visible_segments_on_move = field(
        default=True, type=bool, kw_only=True, repr=False
    )

    def __attrs_post_init__(self):
        if self.source is not None:
            super().__attrs_post_init__()
        if self.name is None:
            self.name = "anno"
        if self.tags is None:
            self.tags = []

    def _to_neuroglancer_layer_local(self) -> viewer_state.LocalAnnotationLayer:
        if len(self.tags) > MAX_TAG_COUNT:
            raise ValueError(
                f"Too many tags. Only {MAX_TAG_COUNT} distinct tags are allowed and {len(self.tags)} have been provided."
            )

        tag_map = {t: i for i, t in enumerate(self.tags)}
        props = make_annotation_properties(self.tags, tag_base_number=0)
        bindings = make_bindings(props)
        if not isinstance(self.resolution, CoordSpace):
            self.resolution = CoordSpace(resolution=self.resolution)
        if self.shader is None:
            return viewer_state.LocalAnnotationLayer(
                dimensions=self.resolution.to_neuroglancer(),
                annotation_color=self.color,
                annotations=_handle_annotations(
                    self.annotations, tag_map, self.resolution.resolution
                ),
                linked_segmentation_layer=_handle_linked_segmentation(
                    self.linked_segmentation
                ),
                annotation_properties=props,
                tool_bindings=bindings,
                swap_visible_segments_on_move=self.swap_visible_segments_on_move,
            )
        else:
            return viewer_state.LocalAnnotationLayer(
                dimensions=self.resolution.to_neuroglancer(),
                annotation_color=self.color,
                annotations=_handle_annotations(
                    self.annotations, tag_map, self.resolution.resolution
                ),
                linked_segmentation_layer=_handle_linked_segmentation(
                    self.linked_segmentation
                ),
                shader=self.shader,
                annotation_properties=props,
                tool_bindings=bindings,
                swap_visible_segments_on_move=self.swap_visible_segments_on_move,
            )

    def _to_neuroglancer_layer_cloud(self) -> viewer_state.AnnotationLayer:
        return viewer_state.AnnotationLayer(
            source=source_to_neuroglancer(self.source),
            annotation_color=self.color,
            linked_segmentation_layer=_handle_linked_segmentation(
                self.linked_segmentation
            ),
            filter_by_segmentation=_handle_filter_by_segmentation(
                self.filter_by_segmentation,
                self.linked_segmentation,
            ),
            shader=self.shader,
            swap_visible_segments_on_move=self.swap_visible_segments_on_move,
        )

    def to_neuroglancer_layer(
        self,
    ) -> Union[viewer_state.AnnotationLayer, viewer_state.LocalAnnotationLayer]:
        super().to_neuroglancer_layer()
        if self.source is None:
            return self._to_neuroglancer_layer_local()
        else:
            return self._to_neuroglancer_layer_cloud()

    def apply_to_neuroglancer(
        self,
        viewer,
    ):
        "Can be a viewer or a viewer.txn()-context state"
        if self.source is None:
            if self.resolution is None:
                raise ValueError(
                    f"Resolution for annotation layer '{self.name}' must be set before converting to Neuroglancer"
                )
            if self.linked_segmentation is True:
                for l in viewer.layers:
                    if l.type == "segmentation":
                        self.linked_segmentation = l.name
                        break
                else:
                    raise ValueError(
                        "No SegmentationLayer found in viewer to link to. Please set linked_segmentation manually."
                    )
            if self.set_position:
                if len(self.annotations) > 0:
                    if not isinstance(self.resolution, CoordSpace):
                        self.resolution = CoordSpace(resolution=self.resolution)
                    first_anno = self.annotations[0]
                    scale = viewer.dimensions.scales * 10**9
                    if isinstance(first_anno, PointAnnotation):
                        new_position = _handle_annotations([first_anno], {}, scale)[
                            0
                        ].point
                    elif isinstance(first_anno, LineAnnotation) or isinstance(
                        first_anno, BoundingBoxAnnotation
                    ):
                        new_position = _handle_annotations(
                            [self.annotations[0]],
                            {},
                            scale,
                        )[0].point_a
                    elif isinstance(self.annotations[0], EllipsoidAnnotation):
                        new_position = _handle_annotations(
                            [self.annotations[0]],
                            {},
                            scale,
                        )[0].center
                    viewer.position = new_position
        self._apply_to_neuroglancer(viewer)

    def set_linked_segmentation(
        self,
        layer: Union[str, SegmentationLayer],
    ) -> Self:
        """Add a linked segmentation layer to the annotation layer.

        Parameters
        ----------
        layer : str or SegmentationLayer
            The linked segmentation layer or the string value of its name.

        """
        if isinstance(layer, SegmentationLayer):
            layer = layer.name
        self.linked_segmentation = layer
        return self

    def add_annotations(
        self,
        annotations: list,
    ) -> Self:
        """Add annotations to the layer.

        Parameters
        ----------
        annotations : list
            The annotations to add.

        Returns
        -------
        self
            The AnnotationLayer object with added annotations.
        """
        if self.annotations is None:
            self.annotations = []
        for annotation in annotations:
            if issubclass(type(annotation), AnnotationBase):
                self.annotations.append(annotation)
            else:
                raise ValueError(
                    "Invalid annotation type. Must be a PointAnnotation, LineAnnotation, Ellipse/SphereAnnotation, or BoundingBoxAnnotation."
                )
        return self

    def _handle_annotation_details(
        self,
        df,
        segment_column,
        description_column,
        tag_column,
        tag_bools,
        n_points,
    ):
        if segment_column is not None:
            if isinstance(segment_column, str):
                segments = df[segment_column].values
            else:
                segment_column = segment_column
        else:
            segments = [None] * n_points
        if description_column is not None:
            if isinstance(description_column, str):
                descriptions = df[description_column].values
            else:
                description_column = description_column
        else:
            descriptions = [None] * n_points

        # Handle tags for annotations
        tag_list = [[] for _ in range(n_points)]
        if tag_column is not None:
            if isinstance(tag_column, str):
                for ii, t in enumerate(df[tag_column].values):
                    tag_list[ii].extend([t])
            else:
                for ii, t in enumerate(tag_column):
                    tag_list[ii].extend([t])
        if tag_bools is not None:
            for tag_ in tag_bools:
                row_to_add = np.arange(n_points)[df[tag_]]
                for i in row_to_add:
                    tag_list[i].extend([tag_])

        # Update annotation layer tag list with new unique values
        all_tags = []
        if tag_column is not None:
            if isinstance(tag_column, str):
                all_tags.extend(df[tag_column].unique().tolist())
            else:
                all_tags.extend(np.unique(tag_column).tolist())
        if tag_bools is not None:
            all_tags.extend(tag_bools)
        self.tags.extend(sorted(set([t for t in all_tags if t not in self.tags])))

        return segments, descriptions, tag_list

    def add_points(
        self,
        data: Union[pd.DataFrame, np.ndarray, DataMap],
        point_column: Optional[str] = None,
        segment_column: Optional[str] = None,
        description_column: Optional[str] = None,
        tag_column: Optional[str] = None,
        tag_bools: Optional[list] = None,
        data_resolution: Optional[list] = None,
    ) -> Self:
        """Add point annotations to the layer.

        Parameters
        ----------
        data : pd.DataFrame or np.ndarray
            The DataFrame containing the point annotations or an Nx3 array.
            If using an array, there is no way to set segment IDs, descriptions, or tags.
        point_column : str
            The column name for the point coordinates.
            Can be a prefix for a split column name with _x, _y, _z as suffixes.
        segment_column : str, optional
            The column name for the segment IDs. Default is None.
        description_column : str, optional
            The column name for the segment descriptions. Default is None.
        tag_column : str, optional
            The column name for segment descripion. Default is None.
        tag_column: str, optional
            The column name for the tags. Default is None.
            Row values should be strings. Nones and NAs will be skipped.
        tag_bools: list, optional
            List of column names to treat as a tag, with values being booleans
            indicating whether the tag is present or not. Default is None.
        data_resolution : list, optional
            The resolution of the data. If None, follows from layer resolution.
            Default is None.

        Returns
        -------
        SegmentationLayer
            The SegmentationLayer object with added point annotations.
        """
        if isinstance(data, DataMap):
            self._register_datamap(
                key=data,
                func=self.add_points,
                point_column=point_column,
                segment_column=segment_column,
                description_column=description_column,
                tag_column=tag_column,
                tag_bools=tag_bools,
                data_resolution=data_resolution,
            )
            return self
        if isinstance(data, pd.DataFrame) or data is None:
            if isinstance(point_column, str):
                point_column = split_point_columns(point_column, data.columns)
                points = data[point_column].values
            else:
                points = np.array(point_column).reshape(-1, 3)
        elif is_list_like(data):
            points = np.array(data).reshape(-1, 3)
        else:
            raise ValueError("Invalid format for data")

        segments, descriptions, tag_list = self._handle_annotation_details(
            data,
            segment_column,
            description_column,
            tag_column,
            tag_bools,
            len(points),
        )

        self.add_annotations(
            [
                PointAnnotation(
                    point=p,
                    segments=seg,
                    description=d,
                    tags=t,
                    resolution=data_resolution,
                )
                for p, seg, d, t in zip(points, segments, descriptions, tag_list)
            ]
        )
        return self

    def add_lines(
        self,
        data: pd.DataFrame,
        point_a_column: str,
        point_b_column: str,
        segment_column: Optional[str] = None,
        description_column: Optional[str] = None,
        tag_column: Optional[str] = None,
        tag_bools: Optional[list] = None,
        data_resolution: Optional[list] = None,
    ) -> Self:
        """Add point annotations to the layer.

        Parameters
        ----------
        data : pd.DataFrame
            The DataFrame containing the point annotations.
        point_a_column : str
            The column name for the start point coordinates.
            Can be a prefix for a split column name with _x, _y, _z as suffixes.
        point_b_column : str
            The column name for the end point coordinates.
            Can be a prefix for a split column name with _x, _y, _z as suffixes.
        segment_column : str, optional
            The column name for the segment IDs. Default is None.
        description_column : str, optional
            The column name for the segment descriptions. Default is None.
        data_resolution : list, optional
            The resolution of the data. Default is None.
        tag_column: str, optional
            The column name for the tags. Default is None.
            Row values should be strings. Nones and NAs will be skipped.
        tag_bools: list, optional
            List of column names to treat as a tag, with values being booleans
            indicating whether the tag is present or not. Default is None.
        data_resolution : list, optional
            The resolution of the data. If None, follows from layer resolution.
            Default is None.

        Returns
        -------
        SegmentationLayer
            The SegmentationLayer object with added point annotations.
        """
        if isinstance(data, DataMap):
            self._register_datamap(
                key=data,
                func=self.add_lines,
                point_a_column=point_a_column,
                point_b_column=point_b_column,
                segment_column=segment_column,
                description_column=description_column,
                tag_column=tag_column,
                tag_bools=tag_bools,
                data_resolution=data_resolution,
            )
            return self
        if isinstance(data, pd.DataFrame) or data is None:
            if isinstance(point_a_column, str):
                point_a_column = split_point_columns(point_a_column, data.columns)
                points_a = data[point_a_column].values
            else:
                points_a = np.array(point_a_column).reshape(-1, 3)

            if isinstance(point_b_column, str):
                point_b_column = split_point_columns(point_b_column, data.columns)
                points_b = data[point_b_column].values
            else:
                points_b = np.array(point_b_column).reshape(-1, 3)
        else:
            data_a, data_b = data
            points_a = np.array(data_a).reshape(-1, 3)
            points_b = np.array(data_b).reshape(-1, 3)
        if len(points_a) != len(points_b):
            raise ValueError("Point A and Point B arrays must have the same length.")

        segments, descriptions, tag_list = self._handle_annotation_details(
            data,
            segment_column,
            description_column,
            tag_column,
            tag_bools,
            len(points_a),
        )

        self.add_annotations(
            [
                LineAnnotation(
                    pointA=p_a,
                    pointB=p_b,
                    segments=seg,
                    description=d,
                    tags=t,
                    resolution=data_resolution,
                )
                for p_a, p_b, seg, d, t in zip(
                    points_a, points_b, segments, descriptions, tag_list
                )
            ]
        )
        return self

    def add_ellipsoids(
        self,
        data: Union[pd.DataFrame, DataMap],
        center_column: str,
        radii_column: str,
        segment_column: Optional[str] = None,
        description_column: Optional[str] = None,
        tag_column: Optional[str] = None,
        tag_bools: Optional[list] = None,
        data_resolution: Optional[list] = None,
    ) -> Self:
        """Add point annotations to the layer.

        Parameters
        ----------
        data : pd.DataFrame
            The DataFrame containing the point annotations.
        center_column : str
            The column name for the center point coordinates.
            Can be a prefix for a split column name with _x, _y, _z as suffixes.
        radii_column : str
            The column name for the radius values.
            Can be a prefix for a split column name with _x, _y, _z as suffixes.
        segment_column : str, optional
            The column name for the segment IDs. Default is None.
        description_column : str, optional
            The column name for the segment descriptions. Default is None.
        data_resolution : list, optional
            The resolution of the data. Default is None.
        tag_column: str, optional
            The column name for the tags. Default is None.
            Row values should be strings. Nones and NAs will be skipped.
        tag_bools: list, optional
            List of column names to treat as a tag, with values being booleans
            indicating whether the tag is present or not. Default is None.
        data_resolution : list, optional
            The resolution of the data. If None, follows from layer resolution.
            Default is None.

        Returns
        -------
        SegmentationLayer
            The SegmentationLayer object with added point annotations.
        """
        if isinstance(data, DataMap):
            self._register_datamap(
                key=data,
                func=self.add_ellipsoids,
                center_column=center_column,
                radii_column=radii_column,
                segment_column=segment_column,
                description_column=description_column,
                tag_column=tag_column,
                tag_bools=tag_bools,
                data_resolution=data_resolution,
            )
            return self
        if isinstance(data, pd.DataFrame) or data is None:
            if isinstance(center_column, str):
                center_column = split_point_columns(center_column, data.columns)
                centers = data[center_column].values
            else:
                centers = np.array(center_column).reshape(-1, 3)

            if isinstance(radii_column, str):
                radii_column = split_point_columns(radii_column, data.columns)
                radii_vals = data[radii_column].values
            else:
                radii_vals = np.array(radii_column)

        segments, descriptions, tag_list = self._handle_annotation_details(
            data,
            segment_column,
            description_column,
            tag_column,
            tag_bools,
            len(centers),
        )

        self.add_annotations(
            [
                EllipsoidAnnotation(
                    center=c,
                    radii=r,
                    segments=seg,
                    description=d,
                    tags=t,
                    resolution=data_resolution,
                )
                for c, r, seg, d, t in zip(
                    centers, radii_vals, segments, descriptions, tag_list
                )
            ]
        )
        return self

    def add_boxes(
        self,
        data: pd.DataFrame,
        point_a_column: str,
        point_b_column: str,
        segment_column: Optional[str] = None,
        description_column: Optional[str] = None,
        tag_column: Optional[str] = None,
        tag_bools: Optional[list] = None,
        data_resolution: Optional[list] = None,
    ) -> Self:
        """Add point annotations to the layer.

        Parameters
        ----------
        data : pd.DataFrame
            The DataFrame containing the point annotations.
        point_a_column : str
            The column name for the start point coordinates of the box.
            Can be a prefix for a split column name with _x, _y, _z as suffixes.
        point_b_column : str
            The column name for the end point coordinates of the box
            Can be a prefix for a split column name with _x, _y, _z as suffixes.
        segment_column : str, optional
            The column name for the segment IDs. Default is None.
        description_column : str, optional
            The column name for the segment descriptions. Default is None.
        data_resolution : list, optional
            The resolution of the data. Default is None.
        tag_column: str, optional
            The column name for the tags. Default is None.
            Row values should be strings. Nones and NAs will be skipped.
        tag_bools: list, optional
            List of column names to treat as a tag, with values being booleans
            indicating whether the tag is present or not. Default is None.
        data_resolution : list, optional
            The resolution of the data. If None, follows from layer resolution.
            Default is None.

        Returns
        -------
        SegmentationLayer
            The SegmentationLayer object with added point annotations.
        """
        if isinstance(data, DataMap):
            self._register_datamap(
                key=data,
                func=self.add_boxes,
                point_a_column=point_a_column,
                point_b_column=point_b_column,
                segment_column=segment_column,
                description_column=description_column,
                tag_column=tag_column,
                tag_bools=tag_bools,
                data_resolution=data_resolution,
            )
            return self
        if isinstance(data, pd.DataFrame) or data is None:
            if isinstance(point_a_column, str):
                point_a_column = split_point_columns(point_a_column, data.columns)
                points_a = data[point_a_column].values
            else:
                points_a = np.array(point_a_column).reshape(-1, 3)

            if isinstance(point_b_column, str):
                point_b_column = split_point_columns(point_b_column, data.columns)
                points_b = data[point_b_column].values
            else:
                points_b = np.array(point_b_column).reshape(-1, 3)
        else:
            points_a, points_b = data
            points_a = np.array(points_a).reshape(-1, 3)
            points_b = np.array(points_b).reshape(-1, 3)

        segments, descriptions, tag_list = self._handle_annotation_details(
            data,
            segment_column,
            description_column,
            tag_column,
            tag_bools,
            len(points_a),
        )
        self.add_annotations(
            [
                BoundingBoxAnnotation(
                    point_a=p_a,
                    point_b=p_b,
                    segments=seg,
                    description=d,
                    tags=t,
                    resolution=data_resolution,
                )
                for p_a, p_b, seg, d, t in zip(
                    points_a, points_b, segments, descriptions, tag_list
                )
            ]
        )
        return self

    @property
    def tag_map(self) -> dict:
        return {tag: f"tagTool_{ii}" for ii, tag in enumerate(self.tags)}

    def add_shader(self, shader: str) -> Self:
        """Add a shader to the layer.

        Parameters
        ----------
        shader : str
            The shader to add.

        """
        self.shader = shader
        return self



import numpy as np
import pandas as pd
import pytest

from nglui.segmentprops import SegmentProperties


@pytest.fixture
def test_df():
    return pd.DataFrame(
        {
            "seg_id": np.arange(0, 100),
            "cell_type": 30 * ["ct_a"] + 30 * ["ct_b"] + 40 * ["ct_c"],
            "category": 50 * ["cat_1"] + 40 * ["cat_2"] + 10 * [None],
            "number_int": np.arange(300, 400),
            "number_float": np.arange(300, 400) + 0.1,
            "tag_a": 90 * [False] + 10 * [True],
            "tag_b": 95 * [True] + 5 * [False],
        }
    )


@pytest.fixture
def test_categorical_df():
    df = pd.DataFrame(
        {
            "seg_id": np.arange(0, 100),
            "cell_type": 30 * ["ct_a"] + 30 * ["ct_b"] + 40 * ["ct_c"],
            "category": 50 * ["cat_1"] + 40 * ["cat_2"] + 10 * [None],
            "number_int": np.arange(300, 400),
            "number_float": np.arange(300, 400) + 0.1,
            "tag_a": 90 * [False] + 10 * [True],
            "tag_b": 95 * [True] + 5 * [False],
        }
    )
    df["cell_type"] = df["cell_type"].astype(
        pd.CategoricalDtype(categories=["ct_d", "ct_c", "ct_b", "ct_a"], ordered=True)
    )
    return df


@pytest.fixture
def test_null_df():
    return pd.DataFrame(
        {
            "seg_id": np.arange(0, 100),
            "cell_type": 30 * ["ct_a"] + 30 * [""] + 40 * [None],
            "category": np.nan,
            "number_int": np.arange(300, 400),
            "number_float": np.arange(300, 400) + 0.1,
            "tag_a": 90 * [False] + 10 * [True],
            "tag_b": 50 * [True] + 50 * [False],
        }
    )


@pytest.fixture
def test_all_null_df():
    return pd.DataFrame(
        {
            "seg_id": np.arange(0, 100),
            "cell_type": 30 * [None] + 30 * [""] + 40 * [None],
            "category": np.nan,
            "number_int": np.arange(300, 400),
            "number_float": np.arange(300, 400) + 0.1,
            "tag_a": 90 * [False] + 10 * [False],
            "tag_b": 50 * [False] + 50 * [False],
        }
    )


@pytest.fixture
def test_segprops():
    return {
        "@type": "neuroglancer_segment_properties",
        "inline": {
            "ids": ["44", "70", "77", "38", "68", "2", "19", "41", "67", "3"],
            "properties": [
                {
                    "id": "label",
                    "type": "label",
                    "values": [
                        "44",
                        "70",
                        "77",
                        "38",
                        "68",
                        "2",
                        "19",
                        "41",
                        "67",
                        "3",
                    ],
                },
                {
                    "id": "tags",
                    "type": "tags",
                    "tags": ["ct_a", "ct_b", "ct_c", "tag_a", "tag_b"],
                    "tag_descriptions": [
                        "ct_a",
                        "ct_b",
                        "ct_c",
                        "The first tag",
                        "The second tag",
                    ],
                    "values": [
                        [1, 4],
                        [2, 4],
                        [2, 4],
                        [1, 4],
                        [2, 4],
                        [0, 4],
                        [0, 4],
                        [1, 4],
                        [2, 4],
                        [0, 4],
                    ],
                },
                {
                    "id": "number_int",
                    "type": "number",
                    "values": [344, 370, 377, 338, 368, 302, 319, 341, 367, 303],
                    "data_type": "int32",
                },
                {
                    "id": "number_float",
                    "type": "number",
                    "values": [
                        344.1000061035156,
                        370.1000061035156,
                        377.1000061035156,
                        338.1000061035156,
                        368.1000061035156,
                        302.1000061035156,
                        319.1000061035156,
                        341.1000061035156,
                        367.1000061035156,
                        303.1000061035156,
                    ],
                    "data_type": "float32",
                },
            ],
        },
    }


def test_segment_props(test_df):
    props = SegmentProperties.from_dataframe(
        test_df,
        id_col="seg_id",
        label_col="seg_id",
        number_cols=["number_int", "number_float"],
        tag_value_cols="cell_type",
        tag_bool_cols=["tag_a", "tag_b"],
        tag_descriptions={"tag_a": "The first tag", "tag_b": "The second tag"},
    )

    assert len(props) == 100
    assert len(props.property_description()) == 4
    p_dict = props.to_dict()
    assert p_dict["inline"]["properties"][2]["data_type"] == "int32"
    assert p_dict["inline"]["properties"][1]["tags"][2] == "ct_c"
    rh_props = SegmentProperties.from_dict(p_dict)
    assert len(rh_props) == 100


def test_categorical_props(test_categorical_df):
    props = SegmentProperties.from_dataframe(
        test_categorical_df,
        id_col="seg_id",
        label_col="seg_id",
        number_cols=["number_int", "number_float"],
        tag_value_cols="cell_type",
        tag_bool_cols=["tag_a", "tag_b"],
        tag_descriptions={"tag_a": "The first tag", "tag_b": "The second tag"},
    )

    assert len(props) == 100
    p_dict = props.to_dict()
    ct_a_idx = p_dict["inline"]["properties"][1]["tags"].index("ct_a")
    assert ct_a_idx in p_dict["inline"]["properties"][1]["values"][0]
    assert "ct_a" in p_dict["inline"]["properties"][1]["tags"]
    assert p_dict["inline"]["properties"][2]["data_type"] == "int32"


def test_prepend_col_name(test_categorical_df):
    props = SegmentProperties.from_dataframe(
        test_categorical_df,
        id_col="seg_id",
        label_col="seg_id",
        number_cols=["number_int", "number_float"],
        tag_value_cols="cell_type",
        tag_bool_cols=["tag_a", "tag_b"],
        prepend_col_name=True,
    )

    assert len(props) == 100
    p_dict = props.to_dict()
    assert "cell_type:ct_a" in p_dict["inline"]["properties"][1]["tags"]


def test_segment_props_nulls(test_null_df):
    props = SegmentProperties.from_dataframe(
        test_null_df,
        id_col="seg_id",
        tag_value_cols="cell_type",
        tag_bool_cols=["tag_a", "tag_b"],
        tag_descriptions={"tag_a": "The first tag", "tag_b": "The second tag"},
    )

    assert tuple(props.tag_properties.tags) == ("ct_a", "tag_a", "tag_b")
    p_dict = props.to_dict()
    assert 2 in p_dict["inline"]["properties"][0]["values"][0]
    assert len(p_dict["inline"]["properties"][0]["values"][50]) == 0


def test_segment_props_all_null(test_all_null_df):
    props = SegmentProperties.from_dataframe(
        test_all_null_df,
        id_col="seg_id",
        tag_value_cols="cell_type",
        tag_bool_cols=["tag_a", "tag_b"],
        tag_descriptions={"tag_a": "The first tag", "tag_b": "The second tag"},
    )

    assert tuple(props.tag_properties.tags) == ("tag_a", "tag_b")
    p_dict = props.to_dict()
    assert len(p_dict["inline"]["properties"][0]["values"][0]) == 0


def test_property_conversion(test_segprops):
    props = SegmentProperties.from_dict(test_segprops)
    assert len(props) == 10
    prop_df = props.to_dataframe()
    assert len(prop_df.columns) == 9


def test_multicolumn_label(test_df):
    props = SegmentProperties.from_dataframe(
        test_df,
        id_col="seg_id",
        label_col=["seg_id", "cell_type", "category"],
    )
    assert len(props) == 100
    p_dict = props.to_dict()
    assert p_dict["inline"]["properties"][0]["values"][1] == "1_ct_a_cat_1"

    props_fmt_map = SegmentProperties.from_dataframe(
        test_df,
        id_col="seg_id",
        label_format_map="{seg_id}_{cell_type}_{category}",
    )

    p_dict_fmt = props_fmt_map.to_dict()
    assert (
        p_dict_fmt["inline"]["properties"][0]["values"][1]
        == p_dict["inline"]["properties"][0]["values"][1]
    )


def test_random_sampler(test_df):
    props = SegmentProperties.from_dataframe(
        test_df,
        id_col="seg_id",
        label_col=["seg_id", "cell_type", "category"],
        random_columns=1,
        random_column_prefix="random_sample_test",
    )
    assert "random_sample_test" in props.to_dataframe().columns

    props = SegmentProperties.from_dataframe(
        test_df,
        id_col="seg_id",
        label_col=["seg_id", "cell_type", "category"],
        random_columns=3,
        random_column_prefix="random_sample_test",
    )
    assert "random_sample_test_2" in props.to_dataframe().columns



from . import helpers, shaders
from .base import *
from .ngl_components import *
from .site_utils import (
    add_neuroglancer_site,
    get_default_neuroglancer_site,
    get_neuroglancer_sites,
    neuroglancer_url,
    set_default_neuroglancer_site,
)
